# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

from __future__ import print_function
from __future__ import unicode_literals
import contextlib
import io
import logging
import re
import sys

from cmakelang import lex
from cmakelang import markup

from cmakelang.common import UserError
from cmakelang.lex import TokenType
from cmakelang.parse.argument_nodes import PositionalGroupNode
from cmakelang.parse.common import FlowType, NodeType, TreeNode
from cmakelang.parse.util import comment_is_tag
from cmakelang.parse import simple_nodes

logger = logging.getLogger(__name__)

BLOCK_TYPES = (NodeType.FLOW_CONTROL, NodeType.BODY, NodeType.COMMENT,
               NodeType.STATEMENT, NodeType.WHITESPACE, NodeType.ONOFFSWITCH,
               NodeType.ATWORDSTATEMENT)
GROUP_TYPES = (NodeType.ARGGROUP, NodeType.KWARGGROUP, NodeType.PARGGROUP,
               NodeType.FLAGGROUP, NodeType.PARENGROUP)
SCALAR_TYPES = (NodeType.FUNNAME, NodeType.ARGUMENT, NodeType.KEYWORD,
                NodeType.FLAG, NodeType.ATWORD)
PAREN_TYPES = (NodeType.LPAREN, NodeType.RPAREN)

MATCH_TYPES = BLOCK_TYPES + GROUP_TYPES + SCALAR_TYPES + PAREN_TYPES


def clamp(value, min_value, max_value):
  """Simple double-ended saturation function."""
  if value > max_value:
    return max_value
  if value < min_value:
    return min_value
  return value


def get_comment_lines(config, node):
  """Given a comment node, iterate through it's tokens and generate a list
  of textual lines."""
  inlines = []
  for token in node.children:
    assert isinstance(token, lex.Token), (
        "Unexpected object as child of comment node")
    if token.type == TokenType.COMMENT:
      inline = token.spelling.strip()

      if (isinstance(node, simple_nodes.CommentNode) and
          node.is_explicit_trailing):
        inlines.append(inline[len(config.markup.explicit_trailing_pattern):])
        continue

      if not config.markup.enable_markup:
        # If markup processing is disabled, preserve the entire line regardless
        # of content
        inlines.append(inline[1:])
      elif inline.startswith('#' * config.markup.hashruler_min_length):
        # If the comment starts with multiple hash chars and it is long enough
        # then treat it as a ruler and preserve it.
        inlines.append(inline[1:])
      else:
        # Otherwise strip off extra hash chars to keep things nice and
        # consistent (i.e. remove acciental `##`` prefixes)
        inlines.append(inline.lstrip('#'))
  return inlines


def format_comment_lines(node, stack_context, line_width):
  """
  Reflow comment lines into the given line width, parsing markup as necessary.
  """
  config = stack_context.config
  inlines = get_comment_lines(config, node)

  if (isinstance(node, simple_nodes.CommentNode) and node.is_explicit_trailing):
    prefix = config.markup.explicit_trailing_pattern
  else:
    prefix = "#"

  if not config.markup.enable_markup:
    return [prefix + line.rstrip() for line in inlines]

  if config.markup.literal_comment_pattern is not None:
    literal_comment_regex = re.compile(config.markup.literal_comment_pattern)
    if literal_comment_regex.match('\n'.join(inlines)):
      return [prefix + line.rstrip() for line in inlines]

  if node.children[0] is stack_context.first_token and (
      config.markup.first_comment_is_literal
      or stack_context.first_token.spelling.startswith("#!")):
    return [prefix + line.rstrip() for line in inlines]

  items = markup.parse(inlines, config)
  markup_lines = markup.format_items(config, max(10, line_width - 2), items)
  return [prefix + (" " * len(line[:1])) + line for line in markup_lines]


def normalize_line_endings(instr):
  """
  Remove trailing whitespace and replace line endings with unix line endings.
  They will be replaced with config.endl during output
  """
  return re.sub('[ \t\f\v]*((\r?\n)|(\r\n?))', '\n', instr)


def count_indentation(linestr):
  """return the number of spaces at the start of the line"""
  for idx, char in enumerate(linestr):
    if char != ' ':
      return idx
  return 0


def replace_with_tabs(content, config):
  """
  Apply the tab policy to `content`. If config.use_tabchars is true, then
  replace spaces at the beginning of every line with tabs according to
  config.tab_size and config.fractional_tab_policy.
  """
  if not config.format.use_tabchars:
    return content

  outlines = []
  for inline in content.split("\n"):
    num_spaces = count_indentation(inline)
    num_tabs, fractional = divmod(num_spaces, config.format.tab_size)
    outline = "\t" * num_tabs
    if fractional:
      if config.format.fractional_tab_policy == "use-space":
        outline += " " * fractional
      elif config.format.fractional_tab_policy == "round-up":
        outline += "\t"
      else:
        raise UserError(
            "Invalid fractional_tab_policy='{}'"
            .format(config.format.fractional_tab_policy))
    outline += inline[num_spaces:]
    outlines.append(outline)
  return "\n".join(outlines)


def need_paren_space(spelling, config):
  """
  Return whether or not we need a space between the statement name and the
  starting parenthesis. This aggregates the logic of the two configuration
  options `separate_ctrl_name_with_space` and `separate_fn_name_with_space`.
  """
  upper = spelling.upper()

  if FlowType.get(upper) is not None:
    return config.format.separate_ctrl_name_with_space
  if upper.startswith('END') and FlowType.get(upper[3:]) is not None:
    return config.format.separate_ctrl_name_with_space
  if upper in ["ELSE", "ELSEIF"]:
    return config.format.separate_ctrl_name_with_space

  return config.format.separate_fn_name_with_space


def is_line_comment(node):
  """
  Return true if the node is a pure parser node holding a line comment (i.e.
  not a bracket comment)
  """
  if isinstance(node, CommentNode):
    node = node.pnode

  if not isinstance(node, TreeNode):
    return False

  if not node.children:
    return False

  return node.children[-1].type == TokenType.COMMENT


def get_pathstr(node_path):
  """Given a list of nodes, construct a path string that can be used to
     identify that node in the tree."""

  pathcopy = []
  for node in node_path:
    if isinstance(node, BodyNode):
      continue
    if isinstance(node, ArgGroupNode):
      continue
    pathcopy.append(node)

  # Construct names
  names = []
  for node in pathcopy:
    if isinstance(node, PargGroupNode):
      count = 0
      # pylint: disable=protected-access
      for sibling in node._parent.children:
        if sibling is node:
          break
        if isinstance(sibling, PargGroupNode):
          count += 1
      name = "{}[{}]".format(node.name, count)
    else:
      name = node.name
    names.append(name)

  return "/".join(names)


class Cursor(object):
  """
  Lightweight class to encode integer positions in a 2d grid.

  * x = row
  * y = cols
  """

  def __init__(self, x, y):
    self.x = x  # row
    self.y = y  # col

  def __add__(self, other):
    """Cursor addition is element-wise (i.e. vector) addition"""
    return Cursor(self.x + other[0], self.y + other[1])

  def __sub__(self, other):
    """Cursor subtraction is element-wise (i.e. vector) subtraction"""
    return Cursor(self.x - other[0], self.y - other[1])

  def __getitem__(self, idx):
    """
    A cursor can be accessed as a two-element array. For a cursor `c`,
    `c[0]` refers to the row (`x`) and `c[1]` refers to the column (`y`).
    """

    if idx == 0:
      return self.x
    if idx == 1:
      return self.y

    raise IndexError('Cursor indices must be 0 or 1')

  def __setitem__(self, idx, value):
    """
    Cursor elements can be assigned as a two-element array. For a cursor `c`,
    `c[0]` refers to the row (`x`) and `c[1]` refers to the column (`y`).
    """

    if idx == 0:
      self.x = value
      return
    if idx == 1:
      self.y = value
      return

    raise IndexError('Cursor indices must be 0 or 1')

  def __repr__(self):
    """
    String representation is like "Cursor(12,34)"
    """
    return "Cursor({},{})".format(self.x, self.y)

  def clone(self):
    """
    Return a new `Cursor` object with the same value as this one.
    """
    return Cursor(*self)


class StackContext(object):
  """
  Aggregate information about the current stack. This object is passed down
  through all of the nested :code:`reflow()` function calls.
  """

  def __init__(self, config, first_token=None):
    self.config = config
    self.node_path = []
    self.first_token = first_token

  @contextlib.contextmanager
  def push_node(self, node):
    """
    Push `node` onto the `node_path` and yield a context manager. Pop `node`
    off of `node_path` when the context manager `__exit__()s`
    """
    self.node_path.append(node)
    yield None
    self.node_path.pop(-1)


class AssertTypeDescriptor(object):
  def __init__(self, assert_type, hidden_name):
    self._assert_type = assert_type
    self._hidden_name = hidden_name

  def __get__(self, obj, objtype):
    return getattr(obj, self._hidden_name, None)

  def __set__(self, obj, value):
    assert isinstance(value, self._assert_type)
    setattr(obj, self._hidden_name, value)


class LayoutNode(object):
  """
  An element in the format/layout tree. The structure of the layout tree
  mirrors that of the parse tree. We could store this info the nodes of the
  parse tree itself but it's a little cleaner to keep the functionality
  separate I think.
  """

  _position = AssertTypeDescriptor(Cursor, "__position")
  _size = AssertTypeDescriptor(Cursor, "__size")

  def __init__(self, pnode):
    self.pnode = pnode
    self._position = Cursor(0, 0)  # NOTE(josh): (row, col)
    self._size = Cursor(0, 0)      # NOTE(josh): (rows, cols)

    # If false, reflow was aborted due to algorithmic constraint: e.g.
    # attempt to horiziontally pack a non-bracket comment, which is
    # syntatically invalid.
    self._reflow_valid = False

    # Set to true if this node is the final node of a statement at
    # the current depth
    self.statement_terminal = False

    self._parent = None
    self._children = []
    self._passno = -1

    # The first column past the last column covered by this node. This is
    # different than just "position" + "size" in the case that the node
    # contains multiline string or bracket arguments
    self._colextent = 0
    self._rowextent = 0

    # Depth of this node in the subtree starting at the statement node.
    # If this node is not in a statement subtree the value is zero.
    self._stmt_depth = 0

    # Depth of the deepest descendant.
    self._subtree_depth = 0

    # The tree structure is locked and immutable. `stmt_depth` and
    # `subtree_depth` have been computed and stored.
    self._locked = False

    # Map subpass number to (passno, wrap) decisions
    self._layout_passes = [(0, False)]

    # Set to true if this node's wrap is activated. Note that _wrap may
    # refer to nesting or vertical layout, depending on what type of node
    # we are. This is value only really needs to be communicated between
    # `reflow()` and `_reflow()` but we store it so that we can render it
    # when viewing the tree for debugging.
    self._wrap = False

    assert isinstance(pnode, TreeNode)

  def _index_in_parent(self):
    for idx, child in enumerate(self._parent.children):
      if child is self:
        return idx
    return -1

  def next_sibling(self):
    if self._parent is None:
      return None
    next_idx = self._index_in_parent() + 1
    if next_idx >= len(self._parent.children):
      return None
    return self._parent.children[next_idx]

  @property
  def name(self):
    """
    The class name of the derived node type.
    """
    return self.__class__.__name__

  @property
  def passno(self):
    """
    The active pass-number which contributed the current layout of the
    subtree rooted at this node.
    """
    return self._passno

  @property
  def colextent(self):
    """
    The column index of the right-most character in the layout of the
    subtree rooted at this node. In other words, the width of the
    bounding box for the subtree rooted at this node.
    """
    return self._colextent

  @property
  def reflow_valid(self):
    """
    A boolean flag indicating whether or not the current layout is accepted.
    If False, then further layout passes are required.
    """
    return self._reflow_valid

  @property
  def position(self):
    """
    A cursor with the (row,col) of the first (i.e. top,left) character in the
    subtree rooted at this node.
    """
    return Cursor(*self._position)

  @property
  def node_type(self):
    """
    Return the `NodeType` of the corresponding parser node that generated
    this layout node.
    """
    return self.pnode.node_type

  # NOTE(josh): making children a property disallows direct assignment
  @property
  def children(self):
    """
    A list of children layout nodes
    """
    return self._children

  @property
  def rowextent(self):
    return self._rowextent

  def __repr__(self):
    boolmap = {True: "T", False: "F"}
    return "{}({}),(passno={},wrap={},ok={}) pos:({},{}) ext:({},{})".format(
        self.__class__.__name__, self.node_type.name,
        self._passno, boolmap[self._wrap], boolmap[self._reflow_valid],
        self.position[0], self.position[1],
        self.rowextent, self.colextent)

  def has_terminal_comment(self):
    """
    Return true if this node has a terminal line comment. In particular, this
    implies that no other node may be packed at the output cursor of this
    node's layout, and a line-wrap is required.
    """
    return False

  def get_depth(self):
    """
    Compute and return the depth of the subtree rooted at this node. The
    depth of the tree is the depth of the deepest (leaf) descendant.
    """

    if self._children:
      return 1 + max(child.get_depth() for child in self._children)
    return 1

  # Lock the tree structure and prevent further updates.
  # Compute `stmt_depth` and `subtree_depth`. Replace the children list with
  # a tuple.
  def lock(self, config, stmt_depth=0):
    """
    Lock the tree structure (topology) and prevent further updates. This is
    mostly for sanity checking. It also computes topological statistics such
    as `stmt_depth` and `subtree_depth`, and replaces the mutable list of
    children with an immuatable tuple.
    """
    self._stmt_depth = stmt_depth
    self._subtree_depth = self.get_depth()
    self._children = tuple(self._children)
    self._locked = True

    for child in self._children:
      # pylint: disable=protected-access
      child._parent = self

    if self.node_type == NodeType.STATEMENT:
      nextdepth = 1
    elif stmt_depth > 0:
      nextdepth = stmt_depth + 1
    else:
      nextdepth = 0

    for child in self._children:
      child.lock(config, nextdepth)

  def _reflow(self, stack_context, cursor, passno):
    """
    Overridden by concrete classes to implement the layout of characters.
    """
    raise NotImplementedError()

  def _validate_layout(
      self, stack_context, start_extent, end_extent):
    """
    Return true if the layout is acceptable according to several checks. For
    example, returns false if the content overflows the columnt limit.
    """

    config = stack_context.config

    # If the bounding box overflows the column limit then the layout is
    # automatically voided
    if end_extent[1] > config.format.linewidth:
      return False

    size = end_extent - start_extent

    if not self._wrap:
      # Regardless of nesting, if the content is wrapped horizontally but it
      # exceeds the configured maximum number of lines we must reject it
      # TODO(josh): figure out how to subtract out any terminal comment
      # contributions to the size, as noted in the algorithm doc.
      if size[0] > config.format.max_lines_hwrap:
        if not isinstance(self, (BodyNode, CommentNode, FlowControlNode)):
          return False

      # Or if this nodepath is marked to always be vertical layout
      pathstr = get_pathstr(stack_context.node_path)
      if pathstr in config.format.always_wrap:
        return False

    return True

  def reflow(self, stack_context, cursor, parent_passno=0):
    """
    (re-)compute the layout of this node under the assumption that it should
    be placed at the given `cursor` on the current `parent_passno`.
    """

    assert self._locked
    assert isinstance(self.pnode, TreeNode)

    self._position = cursor.clone()
    outcursor = None

    layout_passes = \
        stack_context.config.format.layout_passes.get(
            self.__class__.__name__, self._layout_passes)

    with stack_context.push_node(self):
      for passno, wrap in layout_passes:
        if passno > parent_passno:
          break
        self._passno = passno
        self._wrap = wrap
        self._reflow_valid = True
        start_extent = cursor.clone()
        outcursor = self._reflow(
            stack_context, cursor.clone(), passno)
        end_extent = Cursor(outcursor[0], self._colextent)
        self._reflow_valid &= self._validate_layout(
            stack_context, start_extent, end_extent)
        if self._reflow_valid:
          break
    assert outcursor is not None
    return outcursor

  def write(self, config, ctx):
    """
    Output text content given the currently configured layout.
    """
    for child in self._children:
      child.write(config, ctx)

  @staticmethod
  def create(pnode):
    """
    Create a new layout node associated with then given parser node.
    """
    # pylint: disable=too-many-return-statements
    if pnode.node_type in SCALAR_TYPES:
      return ScalarNode(pnode)
    if pnode.node_type == NodeType.ONOFFSWITCH:
      return OnOffSwitchNode(pnode)
    if pnode.node_type == NodeType.STATEMENT:
      return StatementNode(pnode)
    if pnode.node_type == NodeType.ATWORDSTATEMENT:
      return AtWordStatementNode(pnode)
    if pnode.node_type == NodeType.KWARGGROUP:
      return KwargGroupNode(pnode)
    if pnode.node_type == NodeType.ARGGROUP:
      return ArgGroupNode(pnode)
    if pnode.node_type in (NodeType.PARGGROUP,
                           NodeType.FLAGGROUP):
      return PargGroupNode(pnode)
    if pnode.node_type == NodeType.PARENGROUP:
      return ParenGroupNode(pnode)
    if pnode.node_type == NodeType.BODY:
      return BodyNode(pnode)
    if pnode.node_type == NodeType.FLOW_CONTROL:
      return FlowControlNode(pnode)
    if pnode.node_type == NodeType.COMMENT:
      return CommentNode(pnode)
    if pnode.node_type == NodeType.WHITESPACE:
      return WhitespaceNode(pnode)
    if pnode.node_type in PAREN_TYPES:
      return ParenNode(pnode)

    raise RuntimeError("Unexpected node type")


class ParenNode(LayoutNode):
  """Holds parenthesis '(' or ')' for statements or boolean groups."""

  @property
  def name(self):
    return self.node_type.name

  def _reflow(self, stack_context, cursor, passno):
    """There is only one possible layout for this node."""
    self._colextent = cursor[1] + 1
    return cursor + (0, 1)

  def write(self, config, ctx):
    if self.node_type == NodeType.LPAREN:
      ctx.outfile.write_at(self.position, '(')
    elif self.node_type == NodeType.RPAREN:
      ctx.outfile.write_at(self.position, ')')
    else:
      raise ValueError("Unrecognized paren type")


class ScalarNode(LayoutNode):
  """
  Holds scalar tokens such as statement names, parentheses, or keyword or
  positional arguments.
  """

  def has_terminal_comment(self):
    if not self.children:
      return False

    if not self.children[-1].pnode.children:
      return False

    # Note: bracket comments do not count as terminal comments
    return self.children[-1].pnode.children[0].type == TokenType.COMMENT

  def _reflow(self, stack_context, cursor, passno):
    """
    Reflow is pretty trivial for a scalar node. We don't have any choices to
    make, there is only one possible rendering.
    """
    assert self.pnode.children
    token = self.pnode.children[0]
    assert isinstance(token, lex.Token)

    # This might be a multiline string or a multiline bracket argument. In
    # that case we need to normalize line endings and flow each line
    lines = normalize_line_endings(token.spelling).split('\n')
    line = lines.pop(0)
    cursor[1] += len(line)
    self._colextent = cursor[1]

    while lines:
      cursor[0] += 1
      cursor[1] = 0
      line = lines.pop(0)
      cursor[1] += len(line)
      self._colextent = max(self._colextent, cursor[1])

    # Scalar nodes might have terminal comments associated with them. They show
    # up as children in the layout graph. This is the only possible child of
    # a scalar node.
    if self.children:
      # We should not have more than one terminal comment associated with a
      # given scalar node
      assert len(self.children) == 1

      # The only kind of children we store for a scalar node are argument
      # comments.
      child = self.children[0]
      assert child.node_type == NodeType.COMMENT

      # Reflow the comment after the scalar
      cursor = child.reflow(stack_context, cursor + (0, 1), passno)
      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)
    return cursor

  def write(self, config, ctx):
    if not ctx.is_active():
      return

    assert self.pnode.children
    token = self.pnode.children[0]
    assert isinstance(token, lex.Token)

    spelling = normalize_line_endings(token.spelling)
    if self.node_type == NodeType.FUNNAME:
      command_case = config.resolve_for_command(
          token.spelling, "format.command_case")
      if command_case in ("lower", "upper"):
        spelling = getattr(token.spelling, command_case)()
      elif command_case == "canonical":
        if (self._parent.pnode.cmdspec is not None
            and self._parent.pnode.cmdspec.spelling is not None):
          spelling = self._parent.pnode.cmdspec.spelling
        else:
          spelling = config.resolve_for_command(
              token.spelling, "spelling", token.spelling.lower())
      else:
        assert command_case == "unchanged", (
            "Unrecognized command case {}".format(command_case))
    elif (self.node_type in (NodeType.KEYWORD, NodeType.FLAG)
          and config.format.keyword_case in ("lower", "upper")):
      spelling = getattr(token.spelling, config.format.keyword_case)()

    ctx.outfile.write_at(self.position, spelling)
    children = list(self.children)
    if children:
      child = children.pop(0)
      assert child.node_type == NodeType.COMMENT
      child.write(config, ctx)

    assert not children


class OnOffSwitchNode(LayoutNode):
  """
  Holds a special-case line comment token such as ``# cmake-format: off`` or
  ``# cmake-format: on``
  """

  @property
  def name(self):
    return self.pnode.children[0].type.name

  def has_terminal_comment(self):
    return True

  def _reflow(self, stack_context, cursor, passno):
    """
    There is only one possible flow as this is a single-token
    """
    assert self.pnode.children
    token = self.pnode.children[0]
    assert isinstance(token, lex.Token)

    # Canonicalize the output. The input regex may be more generous than this
    # exact string.
    if token.type == TokenType.FORMAT_ON:
      spelling = "# cmake-format: on"
    elif token.type == TokenType.FORMAT_OFF:
      spelling = "# cmake-format: off"

    cursor = cursor + (0, len(spelling))
    self._colextent = cursor[1]
    return cursor

  def write(self, config, ctx):
    assert self.pnode.children
    token = self.pnode.children[0]
    assert isinstance(token, lex.Token)

    if token.type == TokenType.FORMAT_ON:
      spelling = "# cmake-format: on"
      if ctx.offswitch_location is None:
        logging.warning(
            "'#cmake-format: on' with no corresponding 'off' at %d:%d",
            token.begin.line, token.begin.col)
      else:
        ctx.infile.seek(ctx.offswitch_location.offset, 0)
        copy_size = token.begin.offset - ctx.offswitch_location.offset
        copy_bytes = ctx.infile.read(copy_size)
        copy_text = copy_bytes.decode('utf-8')
        ctx.outfile.write(copy_text)
        ctx.offswitch_location = None
        ctx.outfile.forge_cursor(self.position)
    elif token.type == TokenType.FORMAT_OFF:
      spelling = "# cmake-format: off"
      ctx.offswitch_location = token.end

    ctx.outfile.write_at(self.position, spelling)


class StatementNode(LayoutNode):
  """
  Top-level node for a statement.
  """

  def __init__(self, pnode):
    super(StatementNode, self).__init__(pnode)
    self._layout_passes = [
        (0, False),
        (1, True),
        (2, True),
        (3, True),
        (4, True),
        (5, True),
    ]

  def reflow(self, stack_context, cursor, _=0):  # pylint: disable=unused-argument
    return super(StatementNode, self).reflow(
        stack_context, cursor,
        max(passno for passno, _ in self._layout_passes))

  def get_prefix_width(self, config):
    prefix_width = len(self.name) + 1
    if need_paren_space(self.name, config):
      prefix_width += 1  # For the space before the paren
    return prefix_width

  @property
  def name(self):
    return self.children[0].pnode.children[0].spelling.lower()

  def _validate_layout(
      self, stack_context, start_extent, end_extent):
    config = stack_context.config

    # If the bounding box overflows the column limit then the layout is
    # automatically voided
    if end_extent[1] > config.format.linewidth:
      return False

    size = end_extent - start_extent
    if not self._wrap:
      # If the statement spelling is very long and there is enough content that
      # the content is forced to wrap, then we require the statement content to
      # nest.
      if (size[0] > 1 and
          self.get_prefix_width(config) > config.format.max_prefix_chars):
        return False
    return True

  def _reflow(self, stack_context, cursor, passno):
    # pylint: disable=too-many-statements
    config = stack_context.config
    start_cursor = cursor.clone()
    self._colextent = cursor[1]

    # Layout of the statement name is always the same, we just write it out
    # at the current cursor
    children = list(self.children)
    assert children
    child = children.pop(0)
    assert child.node_type == NodeType.FUNNAME

    cursor = child.reflow(stack_context, cursor, passno)
    self._reflow_valid &= child.reflow_valid
    self._colextent = max(self._colextent, child.colextent)

    funname = child
    token = funname.pnode.children[0]

    assert isinstance(token, lex.Token)
    if need_paren_space(token.spelling.lower(), config):
      cursor[1] += 1

    if self.get_prefix_width(config) <= config.format.min_prefix_chars:
      # If the statement or keyword spelling is too short, then nesting doesn't
      # make sense because (nest + tab-width) will take us right back to the
      # same column as without nesting.
      self._wrap = False

    assert children
    child = children.pop(0)
    assert child.node_type == NodeType.LPAREN
    cursor = child.reflow(stack_context, cursor, passno)
    self._reflow_valid &= child.reflow_valid
    self._colextent = max(self._colextent, child.colextent)

    if self._wrap:
      column_cursor = start_cursor + (1, config.format.tab_size)
      cursor = Cursor(*column_cursor)
    else:
      column_cursor = cursor.clone()

    # NOTE(josh): STATEMENTs should have at most one ARGGROUP child between
    # parentheses.
    child = children.pop(0)
    assert child.node_type is NodeType.ARGGROUP, (
        "Expected ARGGROUP node, but got {} at {}"
        .format(child.node_type, child.pnode))
    # Whether or not an ARGGROUP is statement terminal is kind of a weird
    # notion, but as a sentinel for the future "number of open parens" we'll
    # go ahead and mark it.
    child.statement_terminal = True
    cursor = child.reflow(stack_context, cursor, passno)
    self._reflow_valid &= child.reflow_valid
    # We must keep updating the extent after each child because
    # the child might be an argument with a multiline string or a bracket
    # argument... in which case HPACK might actually wrap to a newline.
    self._colextent = max(self._colextent, child.colextent)
    column_cursor[0] = cursor[0]

    assert children
    prev = child
    child = children.pop(0)
    assert child.node_type == NodeType.RPAREN, \
        "Expected RPAREN but got {}".format(child.node_type)

    # NOTE(josh): dangle parens if it wont fit on the current line or
    # if the user has requested us to always do so
    dangle_parens = False
    if config.format.dangle_parens and cursor[0] > start_cursor[0]:
      # If the configuration requests dangling parentheses, then honor that
      # request so long as the statement doesn't fit on a single line
      dangle_parens = True
    elif cursor[1] >= config.format.linewidth:
      # If the child reflow was unable to reserve a column for us to place our
      # parenthesis, then we must dangle it
      dangle_parens = True
      # But we really want to nest first in this case
      if not self._wrap:
        self._reflow_valid = False
    elif prev.has_terminal_comment():
      # If the final token in an argument list is a line comment, then we must
      # dangle the parenthesis, or it will become part of the line comment
      dangle_parens = True

    column_cursor[0] += 1
    if config.format.dangle_align == "prefix":
      dangle_cursor = Cursor(column_cursor[0], start_cursor[1])
    elif config.format.dangle_align == "prefix-indent":
      dangle_cursor = Cursor(
          column_cursor[0], start_cursor[1] + config.format.tab_size)
    elif config.format.dangle_align == "child":
      dangle_cursor = Cursor(*column_cursor)
    else:
      raise ValueError(
          "Unexpected config.format.dangle_align: {}"
          .format(config.format.dangle_align))

    if dangle_parens:
      cursor = dangle_cursor.clone()

    rparen = child
    initial_rparen_cursor = cursor.clone()
    cursor = rparen.reflow(stack_context, initial_rparen_cursor, passno)
    self._reflow_valid &= rparen.reflow_valid
    # NOTE(josh): don't max rparen.colextent here, since we may decide to
    # dangle it below

    # Trailing comment
    if children:
      cursor[1] += 1
      child = children.pop(0)
      assert child.node_type == NodeType.COMMENT, \
          "Expected COMMENT after RPAREN but got {}".format(child.node_type)
      assert not children
      savecursor = cursor.clone()
      cursor = child.reflow(stack_context, cursor, passno)

      # If the statement trailing comment does not fit in the column after the
      # rparen, then dangle the rparen and try again.
      if not dangle_parens and child.colextent > config.format.linewidth:
        cursor = rparen.reflow(stack_context, dangle_cursor, passno)
        self._reflow_valid &= rparen.reflow_valid
        # NOTE(josh): don't max rparen.colextent here, since we may reverse our
        # dangle decision in the next if block
        savecursor = cursor.clone()
        cursor = child.reflow(stack_context, cursor, passno)

      # If the statement trailing comment still does not fit in the current
      # column then just move it to the next line.
      if child.colextent > config.format.linewidth:
        # NOTE(josh): potentially undangle the paren: if the only reason to
        # dangle it was due to the oversized comment line, then we need to
        # undangle it since the oversized comment didn't fit anyway.
        cursor = rparen.reflow(stack_context, initial_rparen_cursor, passno)
        cursor = child.reflow(
            stack_context, Cursor(savecursor[0] + 1, start_cursor[1]),
            passno)

      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)

    self._colextent = max(self._colextent, rparen.colextent)
    return cursor

  def write(self, config, ctx):
    if not ctx.is_active():
      return
    super(StatementNode, self).write(config, ctx)


class AtWordStatementNode(LayoutNode):
  """
  Top-level node for an atword representing a statement. Such as
  @PACKAGE_INIT@
  """

  def __init__(self, pnode):
    super(AtWordStatementNode, self).__init__(pnode)
    self._layout_passes = [
        (0, False),
    ]

  def reflow(self, stack_context, cursor, _=0):  # pylint: disable=unused-argument
    return super(AtWordStatementNode, self).reflow(
        stack_context, cursor,
        max(passno for passno, _ in self._layout_passes))

  def _reflow(self, stack_context, cursor, passno):
    # pylint: disable=too-many-statements
    config = stack_context.config
    start_cursor = cursor.clone()
    self._colextent = cursor[1]

    # Layout of the statement name is always the same, we just write it out
    # at the current cursor
    children = list(self.children)
    assert children
    child = children.pop(0)
    assert child.node_type == NodeType.ATWORD

    cursor = child.reflow(stack_context, cursor, passno)
    self._reflow_valid &= child.reflow_valid
    self._colextent = max(self._colextent, child.colextent)

    # Trailing comment
    if children:
      cursor[1] += 1
      child = children.pop(0)
      assert child.node_type == NodeType.COMMENT, \
          "Expected COMMENT after RPAREN but got {}".format(child.node_type)
      assert not children
      savecursor = cursor.clone()
      cursor = child.reflow(stack_context, cursor, passno)

      # If the statement trailing comment still does not fit in the current
      # column then just move it to the next line.
      if child.colextent > config.format.linewidth:
        # NOTE(josh): potentially undangle the paren: if the only reason to
        # dangle it was due to the oversized comment line, then we need to
        # undangle it since the oversized comment didn't fit anyway.
        cursor = child.reflow(
            stack_context, Cursor(savecursor[0] + 1, start_cursor[1]),
            passno)

      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)

    return cursor

  def write(self, config, ctx):
    if not ctx.is_active():
      return
    super(AtWordStatementNode, self).write(config, ctx)


class KwargGroupNode(LayoutNode):
  """
  A keyword argument group. Contains a keyword, followed by an argument group.
  """

  def __init__(self, pnode):
    super(KwargGroupNode, self).__init__(pnode)
    self._layout_passes = [
        (0, False),
        (1, False),
        (2, False),
        (3, False),
        (4, False),
        (5, True),
    ]

  def has_terminal_comment(self):
    if self.children[-1].node_type is NodeType.COMMENT:
      return True
    return self.children[-1].has_terminal_comment()

  @property
  def name(self):
    return self.children[0].pnode.children[0].spelling.upper()

  def _validate_layout(
      self, stack_context, start_extent, end_extent):
    config = stack_context.config

    # If the bounding box overflows the column limit then the layout is
    # automatically voided
    if end_extent[1] > config.format.linewidth:
      return False

    size = end_extent - start_extent
    if not self._wrap:
      # If the statement spelling is very long and there is enough content that
      # the content is forced to wrap, then we require the statement content to
      # nest.
      if size[0] > 1 and len(self.name) > config.format.max_prefix_chars:
        return False
    return True

  def _reflow(self, stack_context, cursor, passno):
    config = stack_context.config
    start_cursor = cursor.clone()
    self._colextent = cursor[1]

    # Keyword node should have exactly two children, the keyword and the
    # argument group.
    assert len(self.children) <= 2, (
        "Expected at most 2 children in KWargGroup, got {}"
        .format(len(self.children))
    )

    # Layout of the keyword is always the same, we just write it out
    # at the current cursor
    child = self.children[0]
    assert child.node_type == NodeType.KEYWORD

    if len(self.name) <= config.format.min_prefix_chars:
      # If the statement or keyword spelling is too short, then nesting doesn't
      # make sense because (nest + tab-width) will take us right back to the
      # same column as without nesting.
      self._wrap = False

    cursor = child.reflow(stack_context, cursor, passno)
    self._reflow_valid &= child.reflow_valid
    self._colextent = child.colextent

    # NOTE(josh): Empty kwarg. These are rare, and I think they might actually
    # be syntax errors, but we currently support it so let's not limit the
    # possibility for now
    if len(self.children) == 1:
      return cursor

    # keyword = get_normalized_kwarg(child.pnode.children[0])
    if self._wrap:
      column_cursor = Cursor(
          cursor[0] + 1, start_cursor[1] + config.format.tab_size)
    else:
      column_cursor = cursor + (0, 1)

    child = self.children[1]
    cursor = Cursor(*column_cursor)

    if self.statement_terminal:
      child.statement_terminal = True

    cursor = child.reflow(stack_context, cursor, passno)
    self._reflow_valid &= child.reflow_valid
    self._colextent = max(self._colextent, child.colextent)
    column_cursor[0] = cursor[0]
    return cursor


def filename_node_key(layout_node):
  """
  Return the sort key for sortable arguments nodes. This is the
  case-insensitive spelling of the first token in the node.
  """
  return layout_node.pnode.children[0].spelling.lower()


def sort_arguments(children):
  argument_nodes = []
  for child in children:
    if child.pnode.node_type is NodeType.ARGUMENT:
      argument_nodes.append(child)
  argument_nodes = sorted(argument_nodes, key=filename_node_key)
  nodemap = {
      id(node): idx + 1 for idx, node in enumerate(argument_nodes)
  }
  sortlist = []

  keyidx = 0
  posidx = 0
  for child in children:
    if id(child) in nodemap:
      keyidx = nodemap[id(child)]
      posidx = 0

    sortlist.append((keyidx, posidx, child))
    posidx += 1
  return [child for _, _, child in sorted(sortlist)]


def count_arguments(children):
  """
  Count the number of positional arguments (excluding line comments and
  whitespace) within a parg group.
  """
  count = 0
  for child in children:
    if child.node_type is NodeType.COMMENT:
      continue

    count += 1
  return count


class PargGroupNode(LayoutNode):
  """
  A group of positional arguments.
  """

  def __init__(self, pnode):
    super(PargGroupNode, self).__init__(pnode)
    self._max_pargs_hwrap = None
    self._layout_passes = [
        (0, False),
        (1, False),
        (2, False),
        (3, False),
        (4, True),
    ]

  def has_terminal_comment(self):
    if not self.children:
      return False
    if self.children[-1].node_type is NodeType.COMMENT:
      return True
    return self.children[-1].has_terminal_comment()

  def lock(self, config, stmt_depth=0):
    if (config.format.autosort and
        isinstance(self.pnode, PositionalGroupNode) and
        self.pnode.sortable):
      self._children = sort_arguments(self._children)

    super(PargGroupNode, self).lock(config, stmt_depth)

    self._max_pargs_hwrap = config.format.max_pargs_hwrap
    if (isinstance(self.pnode, PositionalGroupNode)
        and self.pnode.spec is not None
        and self.pnode.spec.max_pargs_hwrap is not None):
      self._max_pargs_hwrap = self.pnode.spec.max_pargs_hwrap

  def _reflow(self, stack_context, cursor, passno):
    config = stack_context.config
    children = list(self.children)
    self._colextent = cursor.y

    prev = None
    child = None
    column_cursor = cursor.clone()
    numpargs = count_arguments(children)

    # "COMMAND" arguments are never columnized, since there is no way for us
    # to make that look good
    is_cmdline = "cmdline" in self.pnode.tags
    if is_cmdline:
      self._wrap = False

    rowcount = 0
    while children:
      prev = child
      child = children.pop(0)
      cursor_is_at_column = True

      if prev is None:
        # This is the first child of the arg group so the cursor is already
        # at the right location and theres nothing for us to update
        rowcount = 1
      elif (is_line_comment(prev)
            or prev.has_terminal_comment()
            or self._wrap):
        column_cursor[0] = cursor[0] + 1
        cursor = Cursor(*column_cursor)
        rowcount += 1
      else:
        cursor_is_at_column = False
        cursor[1] += 1

      if self.statement_terminal and not children:
        child.statement_terminal = True

      input_cursor = cursor
      cursor = child.reflow(stack_context, input_cursor, passno)

      if (not cursor_is_at_column) and (not self._wrap):
        # If we are in horizontal wrapping mode, then we need to look at a
        # couple of things thaty may cause us to wrap.
        needs_wrap = False

        # The obvious case is overflow:
        # If the realized extent overflows the column limit then we need to
        # insert a newline and try again
        if child.colextent > config.format.linewidth:
          needs_wrap = True

        # If this is the last node before the closing parenthesis of the
        # statement then we need to extra character (the closing parenthesis)
        # on the last line. If we overlow that extra character slot, then
        # wrap to a new line
        if self.statement_terminal:
          if cursor[1] + 1 > config.format.linewidth:
            needs_wrap = True

        # If the current node is a comment, then we must have parsed it as a
        # line comment (not an argument comment) and so we should preserve
        # it's status as a line comment and not put it after another argument.
        # Note that tag comments are always line comments, and never consumed
        # as argument comments.
        if ((prev and prev.node_type in SCALAR_TYPES) and
            child.node_type is NodeType.COMMENT and
            not child.is_tag()):
          needs_wrap = True

        if needs_wrap:
          rowcount += 1
          column_cursor[0] = input_cursor[0] + 1
          cursor = child.reflow(stack_context, column_cursor, passno)

      # NOTE(josh): we must keep updating the extent after each child because
      # the child might be an argument with a multiline string or a bracket
      # argument... in which case HPACK might actually wrap to a newline.
      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)

    self._rowextent = rowcount
    # NOTE(josh): there is a subtle distinction between invalidating a reflow
    # and forcing mode=vertical. The difference is whether or not a parent
    # node has to advance it's decision state. If we force to vertical at
    # the start of this function, the parent Statement wont nest this
    # ArgGroup. Therefore, we must invalidate here, rather than forcing
    # _vertical above.
    if is_cmdline:
      self._reflow_valid &= (rowcount <= config.format.max_rows_cmdline)
    elif numpargs > self._max_pargs_hwrap:
      self._reflow_valid &= self._wrap

    return cursor


def count_subgroups(children):
  """
  Count the number of positional or kwarg sub groups in an argument group.
  Ignore comments, and assert that no other types of children are found.
  """
  numgroups = 0
  for child in children:
    if child.node_type in (NodeType.KWARGGROUP, NodeType.PARGGROUP,
                           NodeType.PARENGROUP):
      numgroups += 1
    elif child.node_type in (NodeType.COMMENT, NodeType.ONOFFSWITCH):
      continue
    else:
      raise ValueError(
          "Unexpected node type {} as child of ArgGroupNode"
          .format(child.node_type))
  return numgroups


class ArgGroupNode(LayoutNode):
  """
  A group of arguments. This is the single child node of either a
  `StatementNode` or `KwargGroupNode` which then contains any further
  group nodes.
  """

  def __init__(self, pnode):
    super(ArgGroupNode, self).__init__(pnode)
    self._max_subgroups_hwrap = None
    self._layout_passes = [
        (0, False),
        (1, False),
        (2, False),
        (3, False),
        (4, True),
        (5, True),
    ]

  def lock(self, config, stmt_depth=0):
    super(ArgGroupNode, self).lock(config, stmt_depth)
    self._max_subgroups_hwrap = config.format.max_subgroups_hwrap
    if (hasattr(self.pnode, "cmdspec")
        and getattr(self.pnode, "cmdspec") is not None
        and getattr(self.pnode, "cmdspec").max_subgroups_hwrap is not None):
      self._max_subgroups_hwrap = (
          getattr(self.pnode, "cmdspec").max_subgroups_hwrap)

  def has_terminal_comment(self):
    """
    An ArgGroup is a container for one or more PARGGROUP, FLAGGROUP, or
    KWARGGROUP subtrees. Any terminal comment will belong to one of
    it's children.
    """
    return self.children and (self.children[-1].node_type is NodeType.COMMENT or
                              self.children[-1].has_terminal_comment())

  def _reflow(self, stack_context, cursor, passno):
    config = stack_context.config
    children = list(self.children)
    self._colextent = cursor.y

    prev = None
    child = None

    column_cursor = cursor.clone()
    numgroups = count_subgroups(children)

    while children:
      prev = child
      child = children.pop(0)

      if prev is None:
        # This is the first child of the arg group so the cursor is already
        # at the right location and theres nothing for us to update
        is_first_in_row = True
      elif (is_line_comment(prev)
            or prev.has_terminal_comment()
            or self._wrap):
        column_cursor[0] += 1
        cursor = column_cursor.clone()
        is_first_in_row = True
      else:
        cursor[1] += 1
        is_first_in_row = False

      if self.statement_terminal and not children:
        child.statement_terminal = True

      start_cursor = cursor
      cursor = child.reflow(stack_context, cursor, passno)
      if not is_first_in_row and not self._wrap:
        # If we are in horizontal wrapping mode, then we need to check if the
        # child has overflowed the available column width. If so, then we need
        # to wrap to the next line and try again
        needs_wrap = False
        if child.statement_terminal:
          # If this is the last node before the parenthesis then we need to
          # account for one extra character (the closing parenthesis)
          if cursor[1] + 1 > config.format.linewidth:
            needs_wrap = True

        if child.colextent > config.format.linewidth:
          # If the realized extent overflows the column limit then we need to
          # insert a newline and try again
          needs_wrap = True

        if (cursor - start_cursor)[0] > 1:
          # If the argument is wrapped internally (i.e. has more than two lines)
          # then move it to it's own line
          # TODO(josh): Instead of needs_wrap = True unconditionally in this
          # case, let's layout both options and pick which ever one is best
          needs_wrap = True

        if needs_wrap:
          column_cursor[0] += 1
          cursor = Cursor(*column_cursor)
          cursor = child.reflow(stack_context, cursor, passno)

      # NOTE(josh): we must keep updating the extent after each child because
      # the child might be an argument with a multiline string or a bracket
      # argument... in which case HPACK might actually wrap to a newline.
      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)
      column_cursor[0] = cursor[0]

    # NOTE(josh): there is a subtle distinction between invalidating a reflow
    # and forcing mode=vertical. The difference is whether or not a parent
    # node has to advance it's decision state. If we force to vertical at
    # the start of this function, the parent Statement wont nest this
    # ArgGroup. Therefore, we must invalidate here, rather than forcing
    # _vertical above.
    if numgroups > self._max_subgroups_hwrap:
      self._reflow_valid &= self._wrap

    return cursor

  def write(self, config, ctx):
    if not ctx.is_active():
      return
    super(ArgGroupNode, self).write(config, ctx)


class ParenGroupNode(LayoutNode):
  """
  A parenthetical group. According to cmake syntax rules, this necessarily
  implies a boolean logical expression.
  """

  def __init__(self, pnode):
    super(ParenGroupNode, self).__init__(pnode)
    self._layout_passes = [
        (0, False),
        (1, False),
        (2, False),
        (3, False),
        (4, False),
        (5, True),
    ]

  def has_terminal_comment(self):
    children = list(self.children)
    while children and children[0].node_type != NodeType.RPAREN:
      children.pop(0)

    if children:
      children.pop(0)

    return (children
            and children[-1].pnode.children[0].type == TokenType.COMMENT)

  def _reflow(self, stack_context, cursor, passno):
    config = stack_context.config
    children = list(self.children)
    self._colextent = cursor.y

    # ParenGroupNode is composed of:
    #  * a left parenthensis
    #  * an argument group
    #  * a right parenthesis
    #  * an optional trailing comment
    assert len(children) in (3, 4)

    prev = None
    child = None
    # Paren groups do not themselves nest or wrap, they only have one real
    # child: the primary argument group.
    column_cursor = cursor.clone()

    while children:
      prev = child
      child = children.pop(0)

      if prev is None:
        # This is the first child of the arg group so the cursor is already
        # at the right location and theres nothing for us to update
        pass
      elif prev.node_type == NodeType.LPAREN:
        # No space after LParen
        pass
      elif (is_line_comment(prev)
            or prev.has_terminal_comment()
            or self._wrap):
        cursor[1] = column_cursor[1]
        cursor[0] += 1
      elif child.node_type == NodeType.RPAREN:
        # No space before RPAREN
        pass
      else:
        cursor[1] += 1

      if self.statement_terminal and not children:
        child.statement_terminal = True

      cursor = child.reflow(stack_context, cursor, passno)
      if not self._wrap:
        # If we are in horizontal wrapping mode, then we need to check if the
        # child has overflowed the available columnt width. If so, then we need
        # to wrap to the next line and try again

        needs_wrap = False
        if child.statement_terminal:
          # If this is the last node before the parenthesis then we need to
          # account for one extra character (the closing parenthesis)
          if cursor[1] + 1 > config.format.linewidth:
            needs_wrap = True

        if child.colextent > config.format.linewidth:
          # If the realized extent overflows the column limit then we need to
          # insert a newline and try again
          needs_wrap = True

        if not child.reflow_valid:
          needs_wrap = True

        if needs_wrap:
          column_cursor[0] += 1
          cursor = Cursor(*column_cursor)
          cursor = child.reflow(stack_context, cursor, passno)

      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)
      column_cursor[0] = cursor[0]
    return cursor

  def write(self, config, ctx):
    if not ctx.is_active():
      return
    super(ParenGroupNode, self).write(config, ctx)


class BodyNode(LayoutNode):
  """
  Top-level node for a given "scope" depth. This node is the root of a document,
  or the root of any nested statement scopes.
  """

  def _reflow(self, stack_context, cursor, passno):
    """
    Compute the size of a body block
    """
    column_cursor = cursor.clone()
    self._colextent = 0
    for child in self.children:
      cursor = child.reflow(stack_context, column_cursor, passno)
      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)
      column_cursor[0] = cursor[0] + 1

    return cursor


class FlowControlNode(LayoutNode):
  """
  Top-Level node composed of a flow-control statement and it's associated
  `BodyNodes`.
  """

  def _reflow(self, stack_context, cursor, passno):
    """
    Compute the size of a flowcontrol block
    """
    config = stack_context.config
    self._colextent = 0
    column_cursor = cursor.clone()

    children = list(self.children)
    assert children
    child = children.pop(0)
    assert child.node_type == NodeType.STATEMENT
    cursor = child.reflow(stack_context, column_cursor, passno)
    self._reflow_valid &= child.reflow_valid
    self._colextent = max(self._colextent, child.colextent)
    column_cursor[0] = cursor[0] + 1

    assert children
    child = children.pop(0)
    assert child.node_type == NodeType.BODY
    cursor = child.reflow(
        stack_context, column_cursor + (0, config.format.tab_size), passno)
    self._reflow_valid &= child.reflow_valid
    self._colextent = max(self._colextent, child.colextent)
    column_cursor[0] = cursor[0] + 1

    while True:
      assert children
      child = children.pop(0)
      assert child.node_type == NodeType.STATEMENT
      cursor = child.reflow(stack_context, column_cursor, passno)
      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)
      column_cursor[0] = cursor[0] + 1

      if not children:
        break
      child = children.pop(0)
      assert child.node_type == NodeType.BODY
      cursor = child.reflow(stack_context, column_cursor +
                            (0, config.format.tab_size), passno)
      self._reflow_valid &= child.reflow_valid
      self._colextent = max(self._colextent, child.colextent)
      column_cursor[0] = cursor[0] + 1

    return cursor


class CommentNode(LayoutNode):
  """
  A line comment or bracket comment. If parented by a group node then this
  comment acts as an argument. If parented by a scalar node, then this comment
  acts like an argument comment.
  """

  def __init__(self, pnode):
    super(CommentNode, self).__init__(pnode)
    self._lines = []

  def is_tag(self):
    return comment_is_tag(self.pnode.children[0])

  def _reflow(self, stack_context, cursor, passno):
    """
    Compute the size of a comment block
    """
    config = stack_context.config
    if (len(self.pnode.children) == 1
        and isinstance(self.pnode.children[0], lex.Token)
        and self.pnode.children[0].type == TokenType.BRACKET_COMMENT):
      # TODO(josh): cache content and lines and re-use in _write
      content = normalize_line_endings(self.pnode.children[0].spelling)

      # TODO(josh): dedup with ScalarNode logic
      lines = content.split('\n')
      line = lines.pop(0)
      cursor[1] += len(line)
      self._colextent = cursor[1]

      while lines:
        cursor[0] += 1
        cursor[1] = 0
        line = lines.pop(0)
        cursor[1] += len(line)
        self._colextent = max(self._colextent, cursor[1])
      return cursor

    allocation = config.format.linewidth - cursor[1]
    with stack_context.push_node(self):
      self._lines = lines = list(
          format_comment_lines(self.pnode, stack_context, allocation))
    self._colextent = cursor[1] + max(len(line) for line in lines)

    increment = (len(lines) - 1, len(lines[-1]))
    return cursor + increment

  def write(self, config, ctx):
    if not ctx.is_active():
      return

    if (len(self.pnode.children) == 1
        and isinstance(self.pnode.children[0], lex.Token)
        and self.pnode.children[0].type == TokenType.BRACKET_COMMENT):
      content = normalize_line_endings(self.pnode.children[0].spelling)
      ctx.outfile.write_at(self.position, content)
    else:
      for idx, line in enumerate(self._lines):
        ctx.outfile.write_at(self.position + (idx, 0), line)


class WhitespaceNode(LayoutNode):
  """
  A series of newlines
  """

  def _reflow(self, stack_context, cursor, passno):
    """
    Compute the size of a whitespace block
    """
    return cursor.clone()

  def write(self, config, ctx):
    return


def get_scalar_sequence_len(box_children):
  length = 0
  for child in box_children:
    # Child is not a scalar type
    if child.node_type not in SCALAR_TYPES:
      return length
    length += 1
  return length


def create_box_tree(pnode):
  """
  Recursively construct a layout tree from the given parse tree
  """

  layout_root = LayoutNode.create(pnode)
  child_queue = list(pnode.children)

  while child_queue:
    pchild = child_queue.pop(0)
    if not isinstance(pchild, TreeNode):
      continue

    if (pchild.node_type == NodeType.WHITESPACE
        and pchild.count_newlines() < 2):
      continue

    if pchild.node_type in MATCH_TYPES:
      layout_root.children.append(create_box_tree(pchild))
  return layout_root


def layout_tree(parsetree_root, config, linewidth=None, first_token=None):
  """
  Top-level function to construct a layout tree from a parse tree, and then
  iterate through layout passes until the entire tree is satisfactory. Returns
  the root of the layout tree.
  """

  if linewidth is None:
    linewidth = config.format.linewidth

  root_box = create_box_tree(parsetree_root)
  root_box.lock(config)
  stack_context = StackContext(config, first_token)
  root_box.reflow(stack_context, Cursor(0, 0))

  return root_box


def dump_tree(nodes, outfile=None, indent=None):
  """
  Print a tree of node objects for debugging purposes
  """

  if indent is None:
    indent = ''

  if outfile is None:
    outfile = sys.stdout

  for idx, node in enumerate(nodes):
    outfile.write(indent)
    if idx + 1 == len(nodes):
      outfile.write('└─ ')
    else:
      outfile.write('├─ ')

    if sys.version_info[0] < 3:
      outfile.write(repr(node).decode('utf-8'))
    else:
      outfile.write(repr(node))
    outfile.write('\n')

    if not hasattr(node, 'children'):
      continue

    if idx + 1 == len(nodes):
      dump_tree(node.children, outfile, indent + '    ')
    else:
      dump_tree(node.children, outfile, indent + '│   ')


def dump_tree_upto(nodes, history, outfile=None, indent=None):
  """
  Print a tree of node objects for debugging purposes
  """

  if indent is None:
    indent = ''

  if outfile is None:
    outfile = sys.stdout

  for idx, node in enumerate(nodes):
    outfile.write(indent)
    if idx + 1 == len(nodes):
      outfile.write('└─ ')
    else:
      outfile.write('├─ ')

    if history[0] is node:
      # ANSI red
      outfile.write("\u001b[31m")
      if sys.version_info[0] < 3:
        outfile.write(repr(node).decode('utf-8'))
      else:
        outfile.write(repr(node))
      # ANSI reset
      outfile.write("\u001b[0m")
    else:
      if sys.version_info[0] < 3:
        outfile.write(repr(node).decode('utf-8'))
      else:
        outfile.write(repr(node))
    outfile.write('\n')

    if not hasattr(node, 'children'):
      continue

    subhistory = history[1:]
    if not subhistory:
      continue

    if idx + 1 == len(nodes):
      dump_tree_upto(node.children, subhistory, outfile, indent + '    ')
    else:
      dump_tree_upto(node.children, subhistory, outfile, indent + '│   ')


def tree_string(nodes, history=None):
  outfile = io.StringIO()
  if history:
    dump_tree_upto(nodes, history, outfile)
  else:
    dump_tree(nodes, outfile)
  return outfile.getvalue()


def dump_tree_for_test(nodes, outfile=None, indent=None, increment=None):
  """
  Print a tree of node objects for debugging purposes
  """

  if indent is None:
    indent = ''

  if increment is None:
    increment = '  '

  if outfile is None:
    outfile = sys.stdout

  for node in nodes:
    outfile.write(indent)
    outfile.write("({}, {}, {}, {}, {}, [".format(
        node.node_type, node.passno,
        node.position[0], node.position[1], node.colextent))

    if hasattr(node, 'children') and node.children:
      outfile.write('\n')
      dump_tree_for_test(node.children, outfile, indent + increment, increment)
      outfile.write(indent)
    outfile.write("]),\n")


def test_string(nodes, indent=None, increment=None):
  if indent is None:
    indent = ' ' * 0
  if increment is None:
    increment = ' ' * 2

  outfile = io.StringIO()
  dump_tree_for_test(nodes, outfile, indent, increment)
  return outfile.getvalue()


class CursorFile(object):
  def __init__(self, config):
    self._fobj = io.StringIO()
    self._cursor = Cursor(0, 0)
    self._config = config

  @property
  def cursor(self):
    return Cursor(*self._cursor)

  def assert_at(self, cursor):
    assert (self._cursor[0] == cursor[0]
            and self._cursor[1] == cursor[1]), \
        "self._cursor=({},{}), write_at=({}, {}):\n{}".format(
            self._cursor[0], self._cursor[1], cursor[0], cursor[1],
            self.getvalue())

  def assert_lt(self, cursor):
    assert ((self._cursor[0] < cursor[0]) or (
        self._cursor[0] == cursor[0] and self._cursor[1] <= cursor[1])), \
        "self._cursor=({},{}), write_at=({}, {}):\n{}".format(
            self._cursor[0], self._cursor[1], cursor[0], cursor[1],
            self.getvalue().encode('utf-8', errors='replace'))

  def forge_cursor(self, cursor):
    self._cursor = cursor

  def write_at(self, cursor, text):
    if sys.version_info[0] < 3 and isinstance(text, str):
      text = text.decode('utf-8')
    self.assert_lt(cursor)

    rows = (cursor[0] - self._cursor[0])
    if rows:
      self._fobj.write(self._config.format.endl * rows)
      self._cursor[0] += rows
      self._cursor[1] = 0

    cols = (cursor[1] - self._cursor[1])
    if cols:
      self._fobj.write(' ' * cols)
      self._cursor[1] += cols

    lines = text.split('\n')
    line = lines.pop(0)
    self._fobj.write(line)
    self._cursor[1] += len(line)

    while lines:
      self._fobj.write(self._config.format.endl)
      self._cursor[0] += 1
      self._cursor[1] = 0
      line = lines.pop(0)
      self._fobj.write(line)
      self._cursor[1] += len(line)

  def write(self, copy_text):
    if sys.version_info[0] < 3 and isinstance(copy_text, str):
      copy_text = copy_text.decode('utf-8')
    self._fobj.write(copy_text)

    if '\n' not in copy_text:
      self._cursor[1] += len(copy_text)
    else:
      self._cursor[0] += copy_text.count('\n')
      self._cursor[1] = len(copy_text.split('\n')[-1])

  def getvalue(self):
    return self._fobj.getvalue() + self._config.format.endl


class WriteContext(object):
  """
  Global state for the writing functions
  """

  def __init__(self, config, infile_content):
    # The offset within the sourcefile of the end of the offswitch token

    self.offswitch_location = None
    if sys.version_info[0] < 3:
      assert isinstance(infile_content, unicode)

    # NOTE(josh): must be BytesIO because we use byte-offsets for raw
    # transcription (i.e. between cmake-format: off and cmake-format: on)
    self.infile = io.BytesIO(bytearray(infile_content, 'utf-8'))
    self.outfile = CursorFile(config)

  def is_active(self):
    return self.offswitch_location is None


def write_tree(root_box, config, infile_content):
  """
  Format the tree for size only, then print all of the boxes to outfile
  """
  ctx = WriteContext(config, infile_content)
  root_box.write(config, ctx)

  if not ctx.is_active():
    logging.warning("'# cmake-format: off' is never turned back 'on'"
                    "at %d:%d", ctx.offswitch_location.line,
                    ctx.offswitch_location.col)
  return ctx.outfile.getvalue()



import logging

from cmakelang.lex import TokenType
from cmakelang.parse.argument_nodes import (
    ArgGroupNode, PositionalGroupNode, StandardArgTree)
from cmakelang.parse.common import NodeType, TreeNode
from cmakelang.parse.simple_nodes import CommentNode
from cmakelang.parse.util import (
    WHITESPACE_TOKENS,
    get_tag,
    get_normalized_kwarg,
    iter_semantic_tokens,
    only_comments_and_whitespace_remain,
    PositionalSpec,
    should_break
)

logger = logging.getLogger(__name__)


def parse_add_library_standard(ctx, tokens, breakstack, sortable):
  """
  ::

    add_library(<name> [STATIC | SHARED | MODULE]
               [EXCLUDE_FROM_ALL]
               source1 [source2 ...])

  :see: https://cmake.org/cmake/help/v3.0/command/add_library.html
  """
  # pylint: disable=too-many-statements

  parsing_name = 1
  parsing_type = 2
  parsing_flag = 3
  parsing_sources = 4

  tree = ArgGroupNode()

  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens and tokens[0].type in WHITESPACE_TOKENS:
    tree.children.append(tokens.pop(0))
    continue

  state_ = parsing_name
  parg_group = None
  src_group = None
  active_depth = tree

  flags = ("STATIC", "SHARED", "MODULE", "OBJECT")

  while tokens:
    # This parse function breakson the first right paren, since parenthetical
    # groups are not allowed. A parenthesis might exist in a filename, but
    # if so that filename should be quoted so it wont show up as a RIGHT_PAREN
    # token.
    if tokens[0].type is TokenType.RIGHT_PAREN:
      break

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      active_depth.children.append(tokens.pop(0))
      continue

    # If it's a comment token not associated with an argument, then put it
    # directly into the parse tree at the current depth
    if tokens[0].type in (TokenType.COMMENT,
                          TokenType.BRACKET_COMMENT):
      if state_ > parsing_name:
        if get_tag(tokens[0]) in ("unsort", "unsortable"):
          sortable = False
        elif get_tag(tokens[0]) in ("sort", "sortable"):
          sortable = True
      child = TreeNode(NodeType.COMMENT)
      active_depth.children.append(child)
      child.children.append(tokens.pop(0))
      continue

    if state_ is parsing_name:
      token = tokens.pop(0)
      parg_group = PositionalGroupNode()
      parg_group.spec = PositionalSpec("+")
      active_depth = parg_group
      tree.children.append(parg_group)
      child = TreeNode(NodeType.ARGUMENT)
      child.children.append(token)
      CommentNode.consume_trailing(ctx, tokens, child)
      parg_group.children.append(child)
      state_ += 1
    elif state_ is parsing_type:
      if get_normalized_kwarg(tokens[0]) in flags:
        token = tokens.pop(0)
        child = TreeNode(NodeType.FLAG)
        child.children.append(token)
        CommentNode.consume_trailing(ctx, tokens, child)
        parg_group.children.append(child)
      state_ += 1
    elif state_ is parsing_flag:
      if get_normalized_kwarg(tokens[0]) == "EXCLUDE_FROM_ALL":
        token = tokens.pop(0)
        child = TreeNode(NodeType.FLAG)
        child.children.append(token)
        CommentNode.consume_trailing(ctx, tokens, child)
        parg_group.children.append(child)
      state_ += 1
      src_group = PositionalGroupNode(sortable=sortable, tags=["file-list"])
      src_group.spec = PositionalSpec("+")
      active_depth = src_group
      tree.children.append(src_group)
    elif state_ is parsing_sources:
      token = tokens.pop(0)
      child = TreeNode(NodeType.ARGUMENT)
      child.children.append(token)
      CommentNode.consume_trailing(ctx, tokens, child)
      src_group.children.append(child)

      if only_comments_and_whitespace_remain(tokens, breakstack):
        active_depth = tree

  return tree


def parse_add_library_imported(ctx, tokens, breakstack):
  """
  ::
    add_library(<name> <SHARED|STATIC|MODULE|OBJECT|UNKNOWN> IMPORTED
                [GLOBAL])

    :see: https://cmake.org/cmake/help/latest/command/add_library.html
  """
  return StandardArgTree.parse(
      ctx, tokens, npargs='+',
      kwargs={},
      flags=["SHARED", "STATIC", "MODULE", "OBJECT", "UNKOWN", "IMPORTED",
             "GLOBAL"],
      breakstack=breakstack)


def parse_add_library_object(ctx, tokens, breakstack):
  """
  ::
    add_library(<name> OBJECT <src>...)

    :see: https://cmake.org/cmake/help/latest/command/add_library.html#object-libraries
  """
  tree = ArgGroupNode()

  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens and tokens[0].type in WHITESPACE_TOKENS:
    tree.children.append(tokens.pop(0))
    continue

  ntokens = len(tokens)
  subtree = PositionalGroupNode.parse(ctx, tokens, 2, ["OBJECT"], breakstack)
  assert len(tokens) < ntokens
  tree.children.append(subtree)

  while tokens:
    # Break if the next token belongs to a parent parser, i.e. if it
    # matches a keyword argument of something higher in the stack, or if
    # it closes a parent group.
    if should_break(tokens[0], breakstack):
      break

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    # If it's a comment, then add it at the current depth
    if tokens[0].type in (TokenType.COMMENT,
                          TokenType.BRACKET_COMMENT):
      if not get_tag(tokens[0]) in ("sort", "sortable"):
        child = TreeNode(NodeType.COMMENT)
        tree.children.append(child)
        child.children.append(tokens.pop(0))
        continue

    ntokens = len(tokens)
    subtree = PositionalGroupNode.parse(ctx, tokens, '+', [], breakstack)
    assert len(tokens) < ntokens
    tree.children.append(subtree)
  return tree


def parse_add_library_alias(ctx, tokens, breakstack):
  """
  ::
    add_library(<name> ALIAS <target>)

    :see: https://cmake.org/cmake/help/latest/command/add_library.html#alias-libraries
  """
  return StandardArgTree.parse(
      ctx, tokens, npargs=3,
      kwargs={},
      flags=["ALIAS"],
      breakstack=breakstack)


def parse_add_library_interface(ctx, tokens, breakstack):
  """
  ::
    add_library(<name> INTERFACE [IMPORTED [GLOBAL]])

    :see: https://cmake.org/cmake/help/latest/command/add_library.html#interface-libraries
  """
  return StandardArgTree.parse(
      ctx, tokens, npargs='+',
      kwargs={},
      flags=["INTERFACE", "IMPORTED", "GLOBAL"],
      breakstack=breakstack)


def parse_add_library(ctx, tokens, breakstack):
  """
  ``add_library()`` has several forms:

  * normal libraires
  * imported libraries
  * object libraries
  * alias libraries
  * interface libraries

  This function is just the dispatcher

  :see: https://cmake.org/cmake/help/latest/command/add_library.html
  """

  semantic_iter = iter_semantic_tokens(tokens)
  # NOTE(josh): first token is always the name of the library
  _ = next(semantic_iter, None)
  # Second token is usually the descriminator, except for INTERFACE
  second_token = next(semantic_iter, None)

  if second_token is None:
    # All add_library() commands should have at least two arguments
    logger.warning("Invalid add_library() command at %s",
                   tokens[0].get_location())
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  descriminator = second_token.spelling.upper()
  parsemap = {
      "OBJECT": parse_add_library_object,
      "ALIAS": parse_add_library_alias,
      "INTERFACE": parse_add_library_interface,
      "IMPORTED": parse_add_library_imported
  }
  if descriminator in parsemap:
    return parsemap[descriminator](ctx, tokens, breakstack)

  third_token = next(semantic_iter, None)
  if third_token is not None:
    descriminator = third_token.spelling.upper()
    if descriminator == "IMPORTED":
      return parse_add_library_imported(ctx, tokens, breakstack)

  # If the descriminator token might be a variable dereference, then it
  # might be hiding the descriminator... so we shouldn't infer
  # sortability unless it is a word that doesn't match any of the descriminator
  # flags
  sortable = True
  if "${" in second_token.spelling or "${" in third_token.spelling:
    sortable = False
  return parse_add_library_standard(ctx, tokens, breakstack, sortable)


def populate_db(parse_db):
  parse_db["add_library"] = parse_add_library



"""
Update dynamic parts of documentation files.
"""

from __future__ import print_function
from __future__ import unicode_literals

import argparse
import difflib
import io
import os
import re
import sys


def format_directive(content, codetag=None):
  # TODO(josh): change the indentation to 2 for some more compact files
  if codetag is None:
    codetag = ""
  if codetag == "table":
    outlines = [".. table:: ", "    :align: center", ""]
  else:
    outlines = [".. code:: {}".format(codetag), ""]
  for line in content.split("\n"):
    outlines.append(("    " + line).rstrip())
  outlines.append("")
  return "\n".join(outlines)


EXTENSION_TO_TAG = {
    ".txt": "text",
    ".py": "python",
    ".cmake": "cmake"
}


def get_dynamic_text(bitsdir, bittag):
  for filename in os.listdir(bitsdir):
    if not filename.startswith(bittag):
      continue

    inpath = os.path.join(bitsdir, filename)
    with io.open(inpath, "r", encoding="utf-8") as infile:
      content = infile.read()

    extension = filename[len(bittag):]
    if filename.endswith("-table.rst"):
      # TODO(josh): replace the following line with
      # codetag = "table"
      # so that tables are center aligned in the output
      return content
    if filename.endswith(".rst"):
      return content

    codetag = EXTENSION_TO_TAG.get(extension, None)
    return format_directive(content, codetag)
  raise RuntimeError("No dynamic text for bit {}".format(bittag))


def process_file(filepath, nextpath, bitsdir):
  """
  Write a copy of `filepath` to `nextpath` with substitutions from dynamic_text
  """
  tag_pattern = re.compile("^.. dynamic: (.*)-(begin|end)$")
  active_section = None

  with io.open(filepath, "r", encoding="utf-8") as infile:
    with io.open(nextpath, "w", encoding="utf-8") as outfile:
      for line in infile:
        match = tag_pattern.match(line.strip())
        if active_section is None:
          outfile.write(line)

        if match:
          if match.group(2) == "begin":
            active_section = match.group(1)
          elif match.group(2) == "end":
            assert active_section == match.group(1), "Unexpected end tag"
            outfile.write("\n")
            content = get_dynamic_text(bitsdir, active_section)
            outfile.write(content)
            active_section = None
            outfile.write(line)
          else:
            raise RuntimeError("Unexpected tag")


def update_file(filepath, bitsdir):
  """
  Create a copy of the file replacing any dynamic sections with updated
  text. Then replace the original with the updated copy.
  """

  nextpath = filepath + ".next"
  process_file(filepath, nextpath, bitsdir)
  os.rename(nextpath, filepath)


def verify_file(filepath, bitsdir):
  """
  Create a copy of the file replacing any dynamic sections with updated
  text. Then verify that the copy is not different than the original
  """

  nextpath = filepath + ".tmp"
  process_file(filepath, nextpath, bitsdir)
  with io.open(filepath, "r", encoding="utf-8") as infile:
    lines_a = infile.read().split("\n")
  with io.open(nextpath, "r", encoding="utf-8") as infile:
    lines_b = infile.read().split("\n")
  os.unlink(nextpath)
  diff = list(difflib.unified_diff(lines_a, lines_b))
  if diff:
    raise RuntimeError("{} is out of date\n{}"
                       .format(filepath, "\n".join(diff)))


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument("--verify", action="store_true")
  parser.add_argument("--bits")
  parser.add_argument("filepath")
  args = parser.parse_args()

  if args.verify and sys.version_info < (3, 6, 0):
    print("Skipping verification on python < 3.6")
    sys.exit(0)

  if args.verify:
    verify_file(args.filepath, args.bits)
  else:
    update_file(args.filepath, args.bits)


if __name__ == "__main__":
  main()



# -*- coding: utf-8 -*-
# pylint: disable=W0613
from __future__ import print_function
from __future__ import unicode_literals

import logging

from cmakelang import lex
from cmakelang.parse.util import (
    COMMENT_TOKENS, WHITESPACE_TOKENS,
    are_column_aligned, next_is_trailing_comment,
    next_is_explicit_trailing_comment,
    is_valid_trailing_comment
)
from cmakelang.parse.common import NodeType, TreeNode

logger = logging.getLogger(__name__)


class WhitespaceNode(TreeNode):
  """Stores a sequence of whitespace tokens at body scope."""

  def __init__(self):
    super(WhitespaceNode, self).__init__(NodeType.WHITESPACE)

  @classmethod
  def consume(cls, ctx, tokens):
    """
    Consume sequential whitespace, removing tokens from the input list and
    returning a whitespace BlockNode
    """
    node = cls()
    while tokens and tokens[0].type in WHITESPACE_TOKENS:
      node.children.append(tokens.pop(0))
    return node


class CommentNode(TreeNode):
  """Stores a sequence of one or more comment tokens separated by at most
     one newline. These comment tokens together form one semantic comment."""

  def __init__(self):
    super(CommentNode, self).__init__(NodeType.COMMENT)
    self.is_explicit_trailing = False
    self.is_implicit_trailing = False

  def __repr__(self):
    base = super(CommentNode, self).__repr__()
    if self.is_explicit_trailing:
      return base + "(explicit trailing)"
    if self.is_implicit_trailing:
      return base + "(implicit trailing)"

    return base

  @classmethod
  def consume(cls, ctx, tokens):
    """
    Consume sequential comment lines, removing tokens from the input list and
    returning a comment Block
    """
    len_before = len(tokens)

    node = cls()
    if tokens[0].type in (
        lex.TokenType.BRACKET_COMMENT,
        lex.TokenType.FORMAT_OFF,
        lex.TokenType.FORMAT_ON):
      # Bracket comments get their own node because they are capable of
      # globbing up their newlines. Thus they are their own semantic comment,
      # and are not part of any larger semantic structures.
      node.children.append(tokens.pop(0))
      return node

    comment_tokens = []
    while tokens and tokens[0].type in COMMENT_TOKENS:
      # If the next comment token is not column-aligned, then we don't
      # merge it with the previous comment line
      if (comment_tokens and
          not are_column_aligned(comment_tokens[-1], tokens[0])):
        break

      comment_token = tokens.pop(0)
      comment_tokens.append(comment_token)
      node.children.append(comment_token)

      # Multiple comments separated by only one newline are joined together into
      # a single block
      if (len(tokens) > 1
          # pylint: disable=bad-continuation
          and tokens[0].type == lex.TokenType.NEWLINE
          and tokens[1].type in COMMENT_TOKENS):
        node.children.append(tokens.pop(0))

      # Multiple comments separated only by one newline and some whitespace are
      # joined together into a single block
      elif (len(tokens) > 2
            # pylint: disable=bad-continuation
            and tokens[0].type == lex.TokenType.NEWLINE
            and tokens[1].type == lex.TokenType.WHITESPACE
            and tokens[2].type in COMMENT_TOKENS):
        node.children.append(tokens.pop(0))
        node.children.append(tokens.pop(0))

    assert len(tokens) < len_before, \
        "CommentNode.consume didn't consume any tokens"
    return node

  @classmethod
  def consume_explicit_trailing(cls, ctx, tokens, parent):
    """
    Consume sequential comment lines, removing tokens from the input list and
    appending the resulting node as a child to the provided parent
    """

    while tokens and tokens[0].type in (lex.TokenType.WHITESPACE,
                                        lex.TokenType.NEWLINE):
      parent.children.append(tokens.pop(0))

    node = cls()
    node.is_explicit_trailing = True
    parent.children.append(node)
    while tokens and next_is_explicit_trailing_comment(ctx.config, tokens):
      node.children.append(tokens.pop(0))

  @classmethod
  def consume_implicit_trailing(cls, ctx, tokens, parent):
    """
    Consume sequential comment lines, removing tokens from the input list and
    appending the resulting node as a child to the provided parent
    """

    if tokens[0].type == lex.TokenType.WHITESPACE:
      parent.children.append(tokens.pop(0))

    node = cls()
    node.is_implicit_trailing = True
    parent.children.append(node)
    comment_tokens = node.children

    comment_tokens = []
    while tokens and is_valid_trailing_comment(tokens[0]):
      if (comment_tokens and
          not are_column_aligned(comment_tokens[-1], tokens[0])):
        break

      comment_token = tokens.pop(0)
      comment_tokens.append(comment_token)
      node.children.append(comment_token)

      # Multiple comments separated by only one newline are joined together into
      # a single block
      if (len(tokens) > 1 and
          tokens[0].type == lex.TokenType.NEWLINE and
          is_valid_trailing_comment(tokens[1])):
        node.children.append(tokens.pop(0))

      # Multiple comments separated only by one newline and some whitespace are
      # joined together into a single block
      # TODO(josh)[873a811]: maybe match only on comment tokens that start at
      # the same column?
      elif (len(tokens) > 2 and
            tokens[0].type == lex.TokenType.NEWLINE and
            tokens[1].type == lex.TokenType.WHITESPACE and
            is_valid_trailing_comment(tokens[2])):
        node.children.append(tokens.pop(0))
        node.children.append(tokens.pop(0))

  @classmethod
  def consume_trailing(cls, ctx, tokens, parent):
    """
    Consume sequential comment lines, removing tokens from the input list and
    appending the resulting node as a child to the provided parent
    """
    if not next_is_trailing_comment(ctx.config, tokens):
      return None

    if next_is_explicit_trailing_comment(ctx.config, tokens):
      return cls.consume_explicit_trailing(ctx, tokens, parent)

    return cls.consume_implicit_trailing(ctx, tokens, parent)


class OnOffNode(TreeNode):
  """Stores a single comment token signifying cmake-format should enable or
     disable.
  """

  def __init__(self):
    super(OnOffNode, self).__init__(NodeType.ONOFFSWITCH)

  @classmethod
  def consume(cls, ctx, tokens):
    """
    Consume a 'cmake-format: [on|off]' comment
    """

    node = cls()
    node.children.append(tokens.pop(0))
    return node


def consume_whitespace_and_comments(ctx, tokens, tree):
  """Consume any whitespace or comments that occur at the current depth
  """
  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens:
    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    # If it's a comment token not associated with an argument, then put it
    # directly into the parse tree at the current depth
    if tokens[0].type in (lex.TokenType.COMMENT,
                          lex.TokenType.BRACKET_COMMENT):
      child = CommentNode.consume(ctx, tokens)
      tree.children.append(child)
      continue

    # If it's a sentinel comment, then add it at the current depth
    if tokens[0].type in (lex.TokenType.FORMAT_OFF,
                          lex.TokenType.FORMAT_ON):
      tree.children.append(OnOffNode.consume(ctx, tokens))
      continue
    break



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import unittest

from cmakelang.format import __main__
from cmakelang import lex
from cmakelang.lex import TokenType


class TestSpecificLexings(unittest.TestCase):

  def assert_tok_types(self, input_str, expected_types):
    """
    Run the lexer on the input string and assert that the result tokens match
    the expected
    """

    self.assertEqual(expected_types,
                     [tok.type for tok in lex.tokenize(input_str)])

  def test_bracket_arguments(self):
    self.assert_tok_types("foo(bar [=[hello world]=] baz)", [
        TokenType.WORD, TokenType.LEFT_PAREN, TokenType.WORD,
        TokenType.WHITESPACE, TokenType.BRACKET_ARGUMENT, TokenType.WHITESPACE,
        TokenType.WORD, TokenType.RIGHT_PAREN])

  def test_bracket_comments(self):
    self.assert_tok_types("foo(bar #[=[hello world]=] baz)", [
        TokenType.WORD, TokenType.LEFT_PAREN, TokenType.WORD,
        TokenType.WHITESPACE, TokenType.BRACKET_COMMENT, TokenType.WHITESPACE,
        TokenType.WORD, TokenType.RIGHT_PAREN])

    self.assert_tok_types("""\
      #[==[This is a bracket comment at some nested level
      #    it is preserved verbatim, but trailing
      #    whitespace is removed.]==]
      """, [TokenType.WHITESPACE, TokenType.BRACKET_COMMENT, TokenType.NEWLINE,
            TokenType.WHITESPACE])

    self.assert_tok_types("""\
      #[[First bracket comment]]
      # intervening comment
      #[[Second bracket comment]]
      """, [
          TokenType.WHITESPACE, TokenType.BRACKET_COMMENT,
          TokenType.NEWLINE, TokenType.WHITESPACE,
          TokenType.COMMENT, TokenType.NEWLINE, TokenType.WHITESPACE,
          TokenType.BRACKET_COMMENT, TokenType.NEWLINE, TokenType.WHITESPACE
      ])

  def test_string(self):
    self.assert_tok_types("""\
      foo(bar "this is a string")
      """, [TokenType.WHITESPACE, TokenType.WORD, TokenType.LEFT_PAREN,
            TokenType.WORD, TokenType.WHITESPACE, TokenType.QUOTED_LITERAL,
            TokenType.RIGHT_PAREN, TokenType.NEWLINE, TokenType.WHITESPACE])

  def test_string_with_quotes(self):
    self.assert_tok_types(r"""
      "this is a \"string"
      """, [TokenType.NEWLINE, TokenType.WHITESPACE, TokenType.QUOTED_LITERAL,
            TokenType.NEWLINE, TokenType.WHITESPACE])

    self.assert_tok_types(r"""
      'this is a \'string'
      """, [TokenType.NEWLINE, TokenType.WHITESPACE, TokenType.QUOTED_LITERAL,
            TokenType.NEWLINE, TokenType.WHITESPACE])

  def test_complicated_string_with_quotes(self):
    # NOTE(josh): compacted example from bug report
    # https://github.com/cheshirekow/cmakelang/issues/28
    self.assert_tok_types(r"""
      install(CODE "message(\"foo ${bar}/${baz}...\")
        subfun(COMMAND ${WHEEL_COMMAND}
                       ERROR_MESSAGE \"error ${bar}/${baz}\"
               )"
      )
      """, [TokenType.NEWLINE, TokenType.WHITESPACE, TokenType.WORD,
            TokenType.LEFT_PAREN, TokenType.WORD, TokenType.WHITESPACE,
            TokenType.QUOTED_LITERAL, TokenType.NEWLINE, TokenType.WHITESPACE,
            TokenType.RIGHT_PAREN, TokenType.NEWLINE, TokenType.WHITESPACE])

  def test_mixed_whitespace(self):
    """
    Ensure that if a newline is part of a whitespace sequence then it is
    tokenized separately.
    """
    self.assert_tok_types(" \n", [TokenType.WHITESPACE, TokenType.NEWLINE])
    self.assert_tok_types("\t\n", [TokenType.WHITESPACE, TokenType.NEWLINE])
    self.assert_tok_types("\f\n", [TokenType.WHITESPACE, TokenType.NEWLINE])
    self.assert_tok_types("\v\n", [TokenType.WHITESPACE, TokenType.NEWLINE])
    self.assert_tok_types("\r\n", [TokenType.NEWLINE])

  def test_indented_comment(self):
    self.assert_tok_types("""\
      # This multiline-comment should be reflowed
      # into a single comment
      # on one line
      """, [TokenType.WHITESPACE, TokenType.COMMENT, TokenType.NEWLINE,
            TokenType.WHITESPACE, TokenType.COMMENT, TokenType.NEWLINE,
            TokenType.WHITESPACE, TokenType.COMMENT, TokenType.NEWLINE,
            TokenType.WHITESPACE])

  def test_byteorder_mark(self):
    self.assert_tok_types(
        """\ufeffcmake_minimum_required(VERSION 2.8.11)""",
        [TokenType.BYTEORDER_MARK, TokenType.WORD,
         TokenType.LEFT_PAREN, TokenType.WORD, TokenType.WHITESPACE,
         TokenType.UNQUOTED_LITERAL, TokenType.RIGHT_PAREN])

  def test_generator_expression(self):
    self.assert_tok_types(
        r'$<JOIN:${ldgen_libraries},\n>',
        [TokenType.UNQUOTED_LITERAL]
    )

  def test_atword(self):
    self.assert_tok_types(
        r'@PACKAGE_INIT@',
        [TokenType.ATWORD]
    )


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import logging
import os
import unittest
import shutil
import subprocess
import sys
import tempfile
import warnings

# NOTE(josh): FileNotFoundError doesn't exist in python2, so we'll define
# it to be the parent class.
try:
  FileNotFoundError
except NameError:
  # pylint: disable=W0622
  FileNotFoundError = IOError


def get_repo_dir():
  """
  Return the path to the repository root
  """
  thisdir = os.path.dirname(os.path.realpath(__file__))
  parent, _ = os.path.split(thisdir)
  parent, _ = os.path.split(parent)
  return parent


def format_signature(armored_packet):
  """
  Append the common that were stripped out of the signature when stored in the
  database.
  """
  return "\n".join(
      ["-----BEGIN PGP SIGNATURE-----", ""]
      + armored_packet
      + ["-----END PGP SIGNATURE-----"]) + "\n"


def construct_agreement_text(template, dataitem):
  """
  Re-construct the agreement text.
  """
  template = template.replace("{{signer_name}}", dataitem["name"])
  template = template.replace("{{signer_email}}", dataitem["email"])
  return template


class TestContributorAgreements(unittest.TestCase):
  """
  Validate the signature in the contributor database.
  """

  def __init__(self, *args, **kwargs):
    super(TestContributorAgreements, self).__init__(*args, **kwargs)
    self.homedir = None
    self.rmtrees = []

  def setUp(self):
    usual_homedir = os.path.expanduser("~/.gnupg")
    pubring_path = os.path.join(usual_homedir, "pubring.kbx")
    if os.path.exists(pubring_path):
      self.homedir = usual_homedir
    else:
      self.homedir = tempfile.mkdtemp(prefix="gpgtmp_")
      self.rmtrees.append(self.homedir)

  def tearDown(self):
    for dirpath in self.rmtrees:
      for _ in range(3):
        try:
          # NOTE(josh): for some reason we see
          # `FileNotFoundError:
          #   [Errno 2] No such file or directory: 'S.gpg-agent.browser'`
          # My guess is that the file is deleted after the directory scan but
          # before the unlink.
          shutil.rmtree(dirpath)
          break
        except FileNotFoundError:
          continue
      else:
        self.fail("FileNotFoundError after several retries")

  def test_signatures(self):
    """
    Iterate over signatures and verify them.
    """
    gpg_argv = ["gpg", "--homedir", self.homedir]

    if sys.version_info < (3, 0, 0):
      self.skipTest("no pgpy on this python version")

    # TODO(josh): For some reason importing pgpy seems to cause the
    # stderr filedescriptor to leak when we subprocess below. pgpy must be
    # doing some ugly subprocess thing on it's own
    warnings.simplefilter("ignore", ResourceWarning)

    # TODO(josh): pgpy seems to import distutils and the version distributed
    # with virtualenv on this system has an outdated import of `imp` instead
    # of `importlib`.
    with warnings.catch_warnings():
      warnings.simplefilter("ignore", DeprecationWarning)
      import pgpy  # pylint: disable=import-error

    repodir = get_repo_dir()
    with open(os.path.join(
        repodir, "cmakelang/contrib/signature_db.json")) as infile:
      sigdb = json.load(infile)

    # First, get a list of keys already in the local keyring
    known_fingerprints = set()
    proc = subprocess.Popen(
        gpg_argv + ["--fingerprint", "--with-colons"],
        stdout=subprocess.PIPE)
    for line in proc.stdout:
      parts = line.decode("utf-8").split(":")
      if parts[0] == "fpr":
        fingerprint = parts[-2]
        known_fingerprints.add(fingerprint[-16:])
    proc.wait()

    # Now do a scan through the database and get a list of any keys we are
    # missing
    needkeys = []
    for sigentry in sigdb:
      clearsign = format_signature(sigentry["signature"])
      sig = pgpy.PGPSignature()
      sig.parse(clearsign)
      if sig.signer not in known_fingerprints:
        needkeys.append(sig.signer)

    if needkeys:
      # TODO(josh): use SKS pool instead of specific server
      result = subprocess.check_call(
          gpg_argv + ["--keyserver", "keyserver.ubuntu.com", "--recv-keys"]
          + needkeys)
      self.assertEqual(
          result, 0, msg="Failed to fetch all keys from keyserver")

    with open(os.path.join(
        repodir, "cmakelang/contrib/individual_ca.txt")) as infile:
      template = infile.read()

    # Then verify all the signatures
    for sigentry in sigdb:
      clearsign = format_signature(sigentry["signature"])
      document = construct_agreement_text(template, sigentry)
      with tempfile.NamedTemporaryFile(delete=False) as outfile:
        outfile.write(clearsign.encode("utf-8"))
        detached_sig = outfile.name
      with tempfile.NamedTemporaryFile(delete=False) as outfile:
        outfile.write(document.encode("utf-8"))
        document_msg = outfile.name

      with open(os.devnull, "w") as devnull:
        proc = subprocess.Popen(
            gpg_argv + ["--verify", detached_sig, document_msg],
            stderr=devnull)
        result = proc.wait()
      self.assertEqual(
          0, result,
          msg="Failed to verify signature for {}\n\n"
              "See {} and {}".format(
                  sigentry["name"], detached_sig, document_msg))
      os.unlink(detached_sig)
      os.unlink(document_msg)


if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG)
  unittest.main()



from __future__ import unicode_literals
import unittest

from cmakelang.format import __main__
from cmakelang import markup


class MockMarkupConfig(object):
  def __init__(self, fence_pattern, ruler_pattern):
    self.fence_pattern = fence_pattern
    self.ruler_pattern = ruler_pattern


class MockRootConfig(object):
  def __init__(self, fence_pattern, ruler_pattern):
    self.markup = MockMarkupConfig(fence_pattern, ruler_pattern)


class TestSpecificParses(unittest.TestCase):

  def assert_item_types(self, input_str, expected_types, strip_indent=6,
                        config=None):
    """
    Run the lexer on the input string and assert that the result tokens match
    the expected
    """

    lines = [line[strip_indent:] for line in input_str.split('\n')]
    self.assertEqual(expected_types,
                     [item.kind for item in markup.parse(lines, config)])

  def test_paragraphs(self):
    self.assert_item_types("""\
      Hello world

      Hello wold""", [markup.CommentType.PARAGRAPH,
                      markup.CommentType.SEPARATOR,
                      markup.CommentType.PARAGRAPH])

  def test_lists(self):
    self.assert_item_types("""\
      This is a paragraph

      * this is a
      * bulleted list
      * of three items

        1. this is another list
        2. of two items

      This is a paragraph""",
                           [markup.CommentType.PARAGRAPH,
                            markup.CommentType.SEPARATOR,
                            markup.CommentType.BULLET_LIST,
                            markup.CommentType.SEPARATOR,
                            markup.CommentType.ENUM_LIST,
                            markup.CommentType.SEPARATOR,
                            markup.CommentType.PARAGRAPH])

  def test_list_indentation(self):
    lines = [line[6:] for line in """\
      * this is a
      * bulleted list
      * of three items

        * this is another list
        * of two items

          * this is a third list
          * of two items
    """.splitlines()]
    items = markup.parse(lines)
    self.assertEqual(items[0].kind, markup.CommentType.BULLET_LIST)
    self.assertEqual(items[0].indent, 0)
    self.assertEqual(items[2].kind, markup.CommentType.BULLET_LIST)
    self.assertEqual(items[2].indent, 2)
    self.assertEqual(items[4].kind, markup.CommentType.BULLET_LIST)
    self.assertEqual(items[4].indent, 4)

  def test_notes(self):
    self.assert_item_types("""\
      This is a comment
      that should be joined but
      TODO(josh): This todo should not be joined with the previous line.
      NOTE(josh): Also this should not be joined with the todo.
      """, [markup.CommentType.PARAGRAPH,
            markup.CommentType.NOTE,
            markup.CommentType.NOTE,
            markup.CommentType.SEPARATOR])

  def test_rulers(self):
    self.assert_item_types("""\
      --------------------
      This is some
      text that I expect
      to reflow
      --------------------
      """, [markup.CommentType.RULER,
            markup.CommentType.PARAGRAPH,
            markup.CommentType.RULER,
            markup.CommentType.SEPARATOR])

  def test_rulers_break_bullets(self):
    self.assert_item_types("""\
      --------------------
      * Bulleted item
      * Bulttied item
      --------------------
      """, [markup.CommentType.RULER,
            markup.CommentType.BULLET_LIST,
            markup.CommentType.RULER,
            markup.CommentType.SEPARATOR])

  def test_fences(self):
    self.assert_item_types("""\
      ~~~
      this is some
         verbatim text
      that should not
         be changed
      ~~~~~~
      """, [markup.CommentType.FENCE,
            markup.CommentType.VERBATIM,
            markup.CommentType.FENCE,
            markup.CommentType.SEPARATOR])

  def test_custom_fences(self):
    self.assert_item_types("""\
      ###
      this is some
         verbatim text
      that should not
         be changed
      ###
      """, [markup.CommentType.PARAGRAPH,
            markup.CommentType.SEPARATOR])

    config = MockRootConfig(
        fence_pattern=r'^\s*([#`~]{3}[#`~]*)(.*)$',
        ruler_pattern=r'^\s*[^\w\s]{3}.*[^\w\s]{3}$')
    self.assert_item_types(
        """\
      ###
      this is some
         verbatim text
      that should not
         be changed
      ###
      """,
        [markup.CommentType.FENCE,
         markup.CommentType.VERBATIM,
         markup.CommentType.FENCE,
         markup.CommentType.SEPARATOR],
        config=config)


if __name__ == '__main__':
  unittest.main()



# pylint: skip-file
"""
Database of known variable names
NOTE: this file is automatically generated by the script
python -Bm cmakelang.tools.parse_cmake_help variables
"""
import re


PATTERNS = [
  {%-for pattern in patterns%}
    "{{pattern}}",
  {%-endfor%}
]


def stripped_patterns():
  """
  Remove named groups from patterns so we can join them together without
  a name collision.
  """
  regex = re.compile(r"\?P<[\w_]+>")
  return [regex.sub("", pattern) for pattern in PATTERNS]


CASE_SENSITIVE_REGEX = re.compile(
    "|".join(stripped_patterns()))
CASE_INSENSITIVE_REGEX = re.compile(
    "|".join(stripped_patterns()), re.IGNORECASE)



# -*- coding: utf-8 -*-
# pylint: disable=R1708
from __future__ import unicode_literals

import contextlib
import difflib
import functools
import inspect
import io
import os
import re
import sys
import unittest

import six

from cmakelang.format import __main__
from cmakelang import configuration
from cmakelang.format import formatter
from cmakelang import lex
from cmakelang import parse
from cmakelang.parse.common import NodeType
from cmakelang.parse.printer import test_string, tree_string


def strip_indent(content, indent=6):
  """
  Strings used in this file are indented by 6-spaces to keep them readable
  within the python code that they are embedded. Remove those 6-spaces from
  the front of each line before running the tests.
  """

  # NOTE(josh): don't use splitlines() so that we get the same result
  # regardless of windows or unix line endings in content.
  return '\n'.join([line[indent:] for line in content.split('\n')])


def overzip(iterable_a, iterable_b):
  """
  Like itertools.izip but instead if the two lists have different sizes then
  the resulting generator will yield a number of pairs equal to the larger of
  the two inputs (rathe than the smaller). The empty list will be padded with
  None elements.
  """
  iter_a = iter(iterable_a)
  iter_b = iter(iterable_b)

  item_a = next(iter_a, None)
  item_b = next(iter_b, None)

  # NOTE(josh): this only matters when overzipping a parse tree. It's not
  # meaningful for overzipping a layout tree, but it doesn't hurt since
  # lexer tokens don't show up in the layout tree
  while isinstance(item_a, lex.Token):
    item_a = next(iter_a, None)

  while item_a is not None and item_b is not None:
    yield(item_a, item_b)
    item_a = next(iter_a, None)
    while isinstance(item_a, lex.Token):
      item_a = next(iter_a, None)
    item_b = next(iter_b, None)

  while item_a is not None:
    yield(item_a, None)
    item_a = next(iter_a, None)
    while isinstance(item_a, lex.Token):
      item_a = next(iter_a, None)

  while item_b is not None:
    yield(None, item_b)
    item_b = next(iter_b, None)


def replace_camel_to_snake(camel_pair):
  """
  Substituion callback for camel_to_snake
  """
  chars = camel_pair.group(0)
  return "{}_{}".format(chars[0], chars[1])


def camel_to_snake(camelstr):
  """
  Translate a camelCaseString into a snake_case_string
  """
  return re.sub("[a-z][A-Z]", replace_camel_to_snake, camelstr).lower()


def find_sidecar(filepath, cls):
  if filepath is not None:
    if filepath.endswith(".py"):
      return filepath[:-3] + ".cmake"
    return filepath

  thisdir = os.path.dirname(__file__)
  clsname = cls.__name__
  if clsname.startswith("Test"):
    clsname = clsname[4:]
  candidate = camel_to_snake(clsname) + ".cmake"
  candidate = os.path.join(thisdir, candidate)
  if os.path.exists(candidate):
    return candidate
  candidate = camel_to_snake(clsname) + "_tests.cmake"
  candidate = os.path.join(thisdir, candidate)
  if os.path.exists(candidate):
    return candidate
  candidate = inspect.getfile(cls)[:-3] + ".cmake"
  if os.path.exists(candidate):
    return candidate
  return None


def assert_lex(test, input_str, expected_types):
  """
  Run the lexer on the input string and assert that the result tokens match
  the expected
  """
  test.assertEqual(expected_types,
                   [tok.type for tok in lex.tokenize(input_str)])


def assert_parse_tree(test, nodes, tups, tree=None, history=None):
  """
  Check the output tree structure against that of expect_tree: a nested tuple
  tree.
  """

  if tree is None:
    tree = nodes

  if history is None:
    history = []

  for node, tup in overzip(nodes, tups):
    if isinstance(node, lex.Token):
      continue
    message = ("For node {} at\n {} within \n{}. "
               "If this is infact correct, copy-paste this:\n\n{}"
               .format(node, tree_string([node]),
                       tree_string(tree, history),
                       test_string(tree, ' ' * 6, ' ' * 2)))
    test.assertIsNotNone(node, msg="Missing node " + message)
    test.assertIsNotNone(tup, msg="Extra node " + message)
    expect_type, expect_children = tup
    test.assertEqual(node.node_type, expect_type,
                     msg="Expected type={} ".format(expect_type) + message)
    assert_parse_tree(test, node.children, expect_children, tree,
                      history + [node])


def assert_parse(test, input_str, expect_tree):
  """
  Run the parser to get the fst, then compare the result to the types in the
  ``expect_tree`` tuple tree.
  """
  tokens = lex.tokenize(input_str)
  fst_root = parse.parse(tokens, test.parse_ctx)
  assert_parse_tree(test, [fst_root], expect_tree)


def assert_layout_tree(test, nodes, tups, tree=None, history=None):
  if tree is None:
    tree = nodes
  if history is None:
    history = []

  for node, tup in overzip(nodes, tups):
    if isinstance(node, lex.Token):
      continue
    subhistory = history + [node]
    message = (" for node {} at\n {} \n\n\n"
               "If the actual result is expected, then update the test with"
               " this:\n{}"
               .format(node,
                       formatter.tree_string(tree, subhistory),
                       formatter.test_string(tree, ' ' * 6, ' ' * 2)))
    test.assertIsNotNone(node, msg="Missing node" + message)
    test.assertIsNotNone(tup, msg="Extra node" + message)
    if len(tup) == 6:
      ntype, wrap, row, col, colextent, expect_children = tup
      test.assertEqual(node.wrap, wrap,
                       msg="Expected wrap={}".format(wrap) + message)
    else:
      ntype, row, col, colextent, expect_children = tup

    test.assertEqual(node.type, ntype,
                     msg="Expected type={}".format(ntype) + message)
    test.assertEqual(node.position[0], row,
                     msg="Expected row={}".format(row) + message)
    test.assertEqual(node.position[1], col,
                     msg="Expected col={}".format(col) + message)
    test.assertEqual(node.colextent, colextent,
                     msg="Expected colextent={}".format(colextent) + message)
    assert_layout_tree(test, node.children, expect_children, tree, subhistory)


def assert_layout(test, input_str, expect_tree, strip_len=0):
  """
  Run the formatter on the input string and assert that the result matches
  the output string
  """

  input_str = strip_indent(input_str, strip_len)
  tokens = lex.tokenize(input_str)
  parse_tree = parse.parse(tokens, test.parse_db)
  box_tree = formatter.layout_tree(parse_tree, test.config)
  assert_layout_tree(test, [box_tree], expect_tree)


def assert_format(test, input_str, output_str=None, strip_len=0):
  """
  Run the formatter on the input string and assert that the result matches
  the output string
  """
  if output_str is None:
    output_str = input_str

  input_str = strip_indent(input_str, strip_len)
  output_str = strip_indent(output_str, strip_len)

  if sys.version_info[0] < 3:
    assert isinstance(input_str, unicode)

  actual_str, _reflow_valid = __main__.process_file(test.config, input_str)
  delta_lines = list(difflib.unified_diff(output_str.split('\n'),
                                          actual_str.split('\n')))
  delta = '\n'.join(delta_lines[2:])

  if actual_str != output_str:
    message = ('Input text:\n-----------------\n{}\n'
               'Output text:\n-----------------\n{}\n'
               'Expected Output:\n-----------------\n{}\n'
               'Diff:\n-----------------\n{}'
               .format(input_str,
                       actual_str,
                       output_str,
                       delta))
    if sys.version_info[0] < 3:
      message = message.encode('utf-8')
    raise AssertionError(message)


def exec_sidecar(test, body, meta, input_str):
  """
  Assert a formatting and, optionally, a lex, parse, or layout tree.
  """
  if input_str is None:
    input_str = body
  expect_lex = meta.pop("expect_lex", None)
  if expect_lex is not None:
    with test.subTest(phase="lex"):
      assert_lex(test, input_str, expect_lex)
  expect_parse = meta.pop("expect_parse", None)
  if expect_parse is not None:
    with test.subTest(phase="parse"):
      assert_parse(test, input_str, expect_parse)
  expect_layout = meta.pop("expect_layout", None)
  if expect_layout is not None:
    with test.subTest(phase="layout"):
      assert_layout(test, input_str, expect_layout)

  test.config = configuration.Configuration(**meta)
  # TODO(josh): just move this into the configuration for the one test where
  # it's needed.
  test.config.parse.fn_spec.add(
      'foo',
      flags=['BAR', 'BAZ'],
      kwargs={
          "HEADERS": '*',
          "SOURCES": '*',
          "DEPENDS": '*'
      })

  with test.subTest(phase="format"):
    assert_format(test, input_str, body)


class WrapTestWithRunFun(object):
  """
  Given a instance of a bound test-method from a TestCase, wrap that with
  a callable that first calls the method, and then calls `
  test.assertExpectations()`. Which is really just an opaque way of
  automatically calling `test.assertExpectations()` at the end of each test
  method.
  """

  def __init__(self, test_object, bound_method):
    self.test_object = test_object
    self.bound_method = bound_method

  def __call__(self):
    self.bound_method()
    self.test_object.assertExpectations()


def consume_bracket_contents(lineiter, terminator):
  """
  Consume the content of a multiline bracket comment
  """
  linebuf = []
  for _, line in lineiter:
    if line == terminator:
      break
    linebuf.append(line)
  return "\n".join(linebuf)


class SidecarMeta(type):
  """
  Since the unittest framework inspects class members prior to calling
  ``setUpClass`` there does not appear to be any additional hooks that we
  can use to automatically load sidecars. We use a metaclass so that when the
  test fixture class object is instanciated (class is defined) we can load the
  sidecars. This way test methods are loaded before ``unittest`` inspects the
  class.
  """
  def __new__(mcs, name, bases, dct):
    subcls = type.__new__(mcs, name, bases, dct)
    if name not in ("MetaBase", "TestBase"):
      subcls.load_sidecar_tests()
    return subcls


def make_test_fun(test_name, test_body, test_meta, input_str):
  def test_fun(self):
    return exec_sidecar(self, test_body, test_meta, input_str)
  if sys.version_info < (3, 0, 0):
    # In python 2.7 test_name is a unicode object. We need to convert it to
    # a string.
    test_name = test_name.encode("utf-8")
  test_fun.__name__ = test_name
  test_fun.__doc__ = " ".join(test_name.split("_")[1:])
  return test_fun


class TestBase(six.with_metaclass(SidecarMeta, unittest.TestCase)):
  """
  Given a bunch of example usages of a particular command, ensure that they
  lex, parse, layout, and format the same as expected.
  """
  kNumSidecarTests = 0
  kExpectNumSidecarTests = 0

  @classmethod
  def append_sidecar_test(cls, test_name, line_buffer, meta_str, input_str):
    """
    Add a new test loaded from the cmake sidecar file
    """

    # strip extra newlines
    while line_buffer and not line_buffer[-1].strip():
      line_buffer.pop(-1)

    meta = {}
    if meta_str is not None:
      meta["NodeType"] = NodeType
      exec(meta_str, meta)  # pylint: disable=exec-used
      meta.pop("__builtins__")
      meta.pop("NodeType")

    body = "\n".join(line_buffer) + "\n"
    closure = make_test_fun(test_name, body, meta, input_str)
    setattr(cls, "test_" + test_name, closure)
    cls.kNumSidecarTests += 1

  @classmethod
  def load_sidecar_tests(cls, filepath=None):
    cmake_sidecar = find_sidecar(filepath, cls)
    if not cmake_sidecar or not os.path.exists(cmake_sidecar):
      return
    with io.open(cmake_sidecar, "r", encoding="utf-8") as infile:
      lines = infile.read().split("\n")

    test_name = None
    line_buffer = []
    meta_str = None
    input_str = None
    lineiter = enumerate(lines)
    for lineno, line in lineiter:
      if line.startswith("# test: "):
        if line_buffer:
          cls.append_sidecar_test(test_name, line_buffer, meta_str, input_str)
        test_name = line[8:]
        line_buffer = []
        meta_str = None
        input_str = None
      elif line == "#[=[" and test_name and not line_buffer:
        meta_str = consume_bracket_contents(lineiter, "]=]")
      elif line == "#[==[" and test_name and not line_buffer:
        input_str = consume_bracket_contents(lineiter, "]==]")
      elif line.endswith("# end-test"):
        if test_name is None:
          raise ValueError(
              "Malformed sidecar {}:{}".format(cmake_sidecar, lineno))
        cls.append_sidecar_test(test_name, line_buffer, meta_str, input_str)
        test_name = None
        line_buffer = []
        meta_str = None
        input_str = None
      else:
        line_buffer.append(line)

    if line_buffer:
      cls.append_sidecar_test(test_name, line_buffer, meta_str, input_str)

  def test_numsidecar(self):
    """
    Sanity check to makesure all sidecar tests are run.
    """
    self.assertEqual(self.kExpectNumSidecarTests, self.kNumSidecarTests)

  def __init__(self, *args, **kwargs):
    super(TestBase, self).__init__(*args, **kwargs)
    self.config = configuration.Configuration()
    parse_db = parse.funs.get_parse_db()
    self.parse_ctx = parse.ParseContext(parse_db)
    self.source_str = None
    self.expect_lex = None
    self.expect_parse = None
    self.expect_layout = None
    self.expect_format = None

    # NOTE(josh): hacky introspective way of automatically calling
    # assertExpectations() at the end of every test_XXX() function
    for name in dir(self):
      if not name.startswith("test_"):
        continue
      value = getattr(self, name)
      if callable(value):
        setattr(self, name, WrapTestWithRunFun(self, value))

  def setUp(self):
    self.config = configuration.Configuration()
    parse_db = parse.funs.get_parse_db()
    self.parse_ctx = parse.ParseContext(parse_db)
    self.config.parse.fn_spec.add(
        'foo',
        flags=['BAR', 'BAZ'],
        kwargs={
            "HEADERS": '*',
            "SOURCES": '*',
            "DEPENDS": '*'
        })

    self.parse_ctx.parse_db.update(
        parse.funs.get_funtree(self.config.parse.fn_spec))

  @contextlib.contextmanager
  def subTest(self, msg=None, **params):
    # pylint: disable=no-member
    if sys.version_info < (3, 4, 0):
      yield None
    else:
      yield super(TestBase, self).subTest(msg=msg, **params)

  def assertExpectations(self):
    # Empty source_str is shorthand for "assertInvariant"
    if self.source_str is None:
      self.source_str = self.expect_format

    if self.expect_lex is not None:
      with self.subTest(phase="lex"):  # pylint: disable=no-member
        assert_lex(self, self.source_str, self.expect_lex)
    if self.expect_parse is not None:
      with self.subTest(phase="parse"):  # pylint: disable=no-member
        assert_parse(self, self.source_str, self.expect_parse)
    if self.expect_layout is not None:
      with self.subTest(phase="layout"):  # pylint: disable=no-member
        assert_layout(self, self.source_str, self.expect_layout)
    if self.expect_format is not None:
      with self.subTest(phase="format"):  # pylint: disable=no-member
        assert_format(self, self.source_str, self.expect_format)


class TestAddCustomCommand(TestBase):
  """
  Test various examples of add_custom_command()
  """
  kExpectNumSidecarTests = 4


class TestComment(TestBase):
  """
  Test various examples involving comments
  """
  kExpectNumSidecarTests = 5


class TestConditional(TestBase):
  """
  Test various examples of commands that take conditional statements
  """
  kExpectNumSidecarTests = 5


class TestCustomCommand(TestBase):
  """
  Test some custom command usages
  """
  kExpectNumSidecarTests = 6


class TestExport(TestBase):
  """
  Test various examples of export()
  """
  kExpectNumSidecarTests = 1


class TestExternalProject(TestBase):
  """
  Test various examples of ExternalProject_xxx()
  """
  kExpectNumSidecarTests = 5


class TestFile(TestBase):
  """
  Test various examples of the file command
  """
  kExpectNumSidecarTests = 10


class TestForeach(TestBase):
  """
  Test various examples of the foreach() function
  """
  kExpectNumSidecarTests = 5


class TestInstall(TestBase):
  """
  Test various examples of the install command
  """
  kExpectNumSidecarTests = 3


class TestSetTargetProperties(TestBase):
  """
  Test various examples of the install command
  """
  kExpectNumSidecarTests = 1


class TestSet(TestBase):
  """
  Test various examples of the set() function
  """
  kExpectNumSidecarTests = 9



from cmakelang import lex
from cmakelang.parse.argument_nodes import (
    ArgGroupNode,
    PositionalGroupNode
)
from cmakelang.parse.common import (
    TreeNode,
    NodeType,
)
from cmakelang.parse.util import (
    WHITESPACE_TOKENS,
    should_break,
)


def parse_empty(ctx, tokens, breakstack):
  """
  ::

    break()
    continue()
    enable_testing()

  :see: https://cmake.org/cmake/help/latest/command/break.html
  :see: https://cmake.org/cmake/help/latest/command/continue.html
  :see: https://cmake.org/cmake/help/latest/command/enable_testing.html
  :see: https://cmake.org/cmake/help/latest/command/return.html
  """
  tree = ArgGroupNode()

  while tokens:
    # Break if the next token belongs to a parent parser, i.e. if it
    # matches a keyword argument of something higher in the stack, or if
    # it closes a parent group.
    if should_break(tokens[0], breakstack):
      break

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    # If it's a comment, then add it at the current depth
    if tokens[0].type in (lex.TokenType.COMMENT,
                          lex.TokenType.BRACKET_COMMENT):
      child = TreeNode(NodeType.COMMENT)
      tree.children.append(child)
      child.children.append(tokens.pop(0))
      continue

    ctx.lint_ctx.record_lint("E1121", location=tokens[0].get_location())
    ntokens = len(tokens)
    subtree = PositionalGroupNode.parse(ctx, tokens, "+", [], breakstack)
    assert len(tokens) < ntokens
    tree.children.append(subtree)
  return tree


def populate_db(parse_db):
  parse_db["break"] = parse_empty
  parse_db["continue"] = parse_empty
  parse_db["enable_testing"] = parse_empty
  parse_db["return"] = parse_empty



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import logging
import sys

from cmakelang.tools.usage_lexer import TokenType, Token, tokenize

logger = logging.getLogger(__name__)


class ChoiceNode(list):
  def __init__(self, required=False):
    super(ChoiceNode, self).__init__()
    self.required = required
    self.append([])

  def __repr__(self):
    tags = []
    if self.required:
      tags.append("required")
    else:
      tags.append("optional")
    return "node {}".format(tags)


def parse(tokens):
  """
  Parse a token stream into a tree of nodes
  """

  parse_stack = [ChoiceNode(required=True)]
  for token in tokens:
    if token.type is TokenType.WHITESPACE:
      continue

    if token.type is TokenType.LPAREN:
      next_group = ChoiceNode(required=True)
      parse_stack[-1][-1].append(next_group)
      parse_stack.append(next_group)
    elif token.type is TokenType.LSQ_BRACKET:
      next_group = ChoiceNode(required=False)
      parse_stack[-1][-1].append(next_group)
      parse_stack.append(next_group)
    elif token.type is TokenType.LA_BRACKET:
      next_group = ChoiceNode(required=True)
      parse_stack[-1][-1].append(next_group)
      parse_stack.append(next_group)
    elif token.type in (TokenType.RSQ_BRACKET, TokenType.RA_BRACKET,
                        TokenType.RPAREN):
      parse_stack.pop(-1)
    elif token.type is TokenType.PIPE:
      parse_stack[-1].append([])
    else:
      parse_stack[-1][-1].append(token)
  return parse_stack[0]


def dump_tree(nodes, outfile=None, indent=None):
  """
  Print a tree of node objects for debugging purposes
  """

  if indent is None:
    indent = ''

  if outfile is None:
    outfile = sys.stdout

  for idx, node in enumerate(nodes):
    outfile.write(indent)
    if idx + 1 == len(nodes):
      outfile.write('└─ ')
      increment = '    '
    else:
      outfile.write('├─ ')
      increment = '│   '

    if isinstance(node, (Token, ChoiceNode)):
      linestr = repr(node)
    elif isinstance(node, list):
      linestr = "choice {}".format(idx)
    else:
      raise ValueError("Unexpected node type {}".format(type(node)))
    outfile.write(linestr)
    outfile.write('\n')

    if isinstance(node, Token):
      continue

    if isinstance(node, ChoiceNode) and len(node) == 1:
      dump_tree(node[0], outfile, indent + increment)
      continue

    dump_tree(node, outfile, indent + increment)


def main():
  """
  Dump parse tree for debugging
  """
  import argparse
  import io
  import os

  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument('infile')
  args = parser.parse_args()

  infile_arg = args.infile
  if args.infile == "-":
    infile_arg = os.dup(sys.stdin.fileno())

  with io.open(infile_arg, 'r') as infile:
    tokens = tokenize(infile.read())
  parse_tree = parse(tokens)
  dump_tree([parse_tree])


if __name__ == '__main__':
  main()



# -*- coding: utf-8 -*-
"""
Parse cmake listfiles and re-emit them with semantic annotations in HTML.

Some options regarding parsing are configurable by providing a configuration
file. The configuration file format is the same as that used by cmake-format,
and the same file can be used for both programs.

cmake-format can spit out the default configuration for you as starting point
for customization. Run with `--dump-config [yaml|json|python]`.
"""
from __future__ import unicode_literals

import argparse
import io
import logging
import os
import sys

import cmakelang
from cmakelang.format import __main__
from cmakelang import configuration
from cmakelang import lex
from cmakelang import parse
from cmakelang import render


EMBED_TPL = """
<template id="renderedcontent">
{{html_content}}
</template>
<iframe id="renderframe" style="width:100%;"></iframe>
<script type="text/javascript">
  var frame = document.getElementById("renderframe");
  frame.addEventListener("load", function() {
    frame.height = frame.contentWindow.document.body.scrollHeight + 30;
  });
  frame.srcdoc = document.getElementById("renderedcontent").innerHTML;
</script>
"""


def annotate_file(config, infile, outfile, outfmt=None):
  """
  Parse the input cmake file, re-format it, and print to the output file.
  """

  infile_content = infile.read()
  if config.format.line_ending == 'auto':
    detected = __main__.detect_line_endings(infile_content)
    config = config.clone()
    config.format.set_line_ending(detected)
  tokens = lex.tokenize(infile_content)
  parse_db = parse.funs.get_parse_db()
  parse_db.update(parse.funs.get_funtree(config.parse.fn_spec))
  ctx = parse.ParseContext(parse_db)
  parse_tree = parse.parse(tokens, ctx)

  if outfmt == "page":
    html_content = render.get_html(parse_tree, fullpage=True)
    outfile.write(html_content)
    return
  if outfmt == "stub":
    html_content = render.get_html(parse_tree, fullpage=False)
    outfile.write(html_content)
    return
  if outfmt == "iframe":
    html_content = render.get_html(parse_tree, fullpage=True)
    wrap_lines = EMBED_TPL.split("\n")
    for line in wrap_lines[:2]:
      outfile.write(line)
      outfile.write("\n")
    outfile.write(html_content)
    for line in wrap_lines[3:]:
      outfile.write(line)
      outfile.write("\n")
    return

  raise ValueError("Invalid output format: {}".format(outfmt))


USAGE_STRING = """
cmake-annotate [-h]
             [--format {page,stub}]
             [-o OUTFILE_PATH]
             [-c CONFIG_FILE]
             infilepath [infilepath ...]
"""


def setup_argparser(arg_parser):
  """
  Add argparse options to the parser.
  """
  arg_parser.add_argument('-v', '--version', action='version',
                          version=cmakelang.__version__)
  arg_parser.add_argument(
      "-f", "--format", choices=["page", "stub", "iframe"], default="stub",
      help="whether to output a standalone `page` complete with <html></html> "
           "tags, or just the annotated content")

  arg_parser.add_argument('-o', '--outfile-path', default=None,
                          help='Where to write the formatted file. '
                               'Default is stdout.')
  arg_parser.add_argument('-c', '--config-file',
                          help='path to configuration file')
  arg_parser.add_argument('infilepaths', nargs='*')


def main():
  """Parse arguments, open files, start work."""

  # set up main logger, which logs everything. We'll leave this one logging
  # to the console
  logging.basicConfig(level=logging.INFO)
  arg_parser = argparse.ArgumentParser(
      description=__doc__,
      formatter_class=argparse.RawDescriptionHelpFormatter,
      usage=USAGE_STRING)
  setup_argparser(arg_parser)
  args = arg_parser.parse_args()

  assert (len(args.infilepaths) == 1
          or args.outfile_path is None), \
      ("if more than one input file is specified, then annotates must be "
       "written to stdout")

  if args.outfile_path is None:
    args.outfile_path = '-'

  if '-' in args.infilepaths:
    assert len(args.infilepaths) == 1, \
        "You cannot mix stdin as an input with other input files"
    assert args.outfile_path == '-', \
        "If stdin is the input file, then stdout must be the output file"

  argdict = __main__.get_argdict(args)
  output_format = argdict.pop("format")

  for infile_path in args.infilepaths:
    # NOTE(josh): have to load config once for every file, because we may pick
    # up a new config file location for each path
    if infile_path == '-':
      config_dict = __main__.get_config(os.getcwd(), args.config_file)
    else:
      config_dict = __main__.get_config(infile_path, args.config_file)
    config_dict.update(argdict)

    cfg = configuration.Configuration(**config_dict)
    if args.outfile_path == '-':
      # NOTE(josh): The behavior or sys.stdout is different in python2 and
      # python3. sys.stdout is opened in 'w' mode which means that write()
      # takes strings in python2 and python3 and, in particular, in python3
      # it does not take byte arrays. io.StreamWriter will write to
      # it with byte arrays (assuming it was opened with 'wb'). So we use
      # io.open instead of open in this case
      outfile = io.open(
          os.dup(sys.stdout.fileno()),
          mode='w', encoding=cfg.encode.output_encoding, newline='')
    else:
      outfile = io.open(
          args.outfile_path, 'w', encoding=cfg.encode.output_encoding,
          newline='')

    if infile_path == '-':
      infile = io.open(
          os.dup(sys.stdin.fileno()),
          mode='r', encoding=cfg.encode.input_encoding, newline='')
    else:
      infile = io.open(infile_path, 'r', encoding=cfg.encode.input_encoding)

    try:
      with infile:
        annotate_file(cfg, infile, outfile, output_format)
    except:
      sys.stderr.write('While processing {}\n'.format(infile_path))
      raise
    finally:
      outfile.close()

  return 0


if __name__ == '__main__':
  sys.exit(main())



# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import sphinx_rtd_theme

project = 'tangentsky'
thisfile = os.path.realpath(__file__)
rootdir = globals().get(
    "rootdir", os.sep.join(thisfile.split(os.sep)[:-3]))

if os.environ.get("READTHEDOCS") == "True":
  # Do any RTD-specific setup here
  pass

# General information about the project.
docname = project + u'doc'
title = project + ' Documentation'
copyright = u'2018-2019, Josh Bialkowski'  # pylint: disable=W0622
author = u'Josh Bialkowski'

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    # autosectionlabel is pretty broken
    # 'sphinx.ext.autosectionlabel',
    # "readthedocs_ext.readthedocs",
]

autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

mathjax_path = (
    "https://cdnjs.cloudflare.com/ajax/libs/mathjax/"
    "2.7.5/MathJax.js?config=TeX-MML-AM_CHTML")

# Add any paths that contain templates here, relative to this directory.
templates_path = []

source_suffix = ['.rst']
master_doc = 'index'

version = u'0.1.0'
release = u'0.1.0'
language = None
exclude_patterns = [
    '.sphinx_build',
    'Thumbs.db',
    '.DS_Store'
]

pygments_style = 'sphinx'
todo_include_todos = True

html_theme = 'sphinx_rtd_theme'
# html_theme = "alabaster"
numfig = True
math_numfig = True
math_number_all = True
math_eqref_format = "Eq. {number}"

html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_theme_options = {
    "collapse_navigation": True,
    "display_version": True,
    "sticky_navigation": True,
}

html_static_path = ["sphinx-static"]
html_style = "css/cheshire_theme.css"

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'globaltoc.html',  # NOTE(josh): added
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
    ]
}

htmlhelp_basename = project + 'doc'

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
    # Latex figure (float) alignment
    # 'figure_align': 'htbp',
}

latex_documents = [
    (master_doc, project + '.tex', title, author, 'manual'),
]

man_pages = [
    (master_doc, project, title,
     [author], 1)
]

texinfo_documents = [
    (master_doc, project, title,
     author, project, 'One line description of project.',
     'Miscellaneous'),
]

intersphinx_mapping = {'https://docs.python.org/': None}



from cmakelang.parse.additional_nodes import ShellCommandNode
from cmakelang.parse.argument_nodes import (
    PositionalParser, StandardArgTree, StandardParser)


def parse_fetchcontent_declare(ctx, tokens, breakstack):
  """
  ::

    FetchContent_Declare(<name> <contentOptions>...)

  :see: https://cmake.org/cmake/help/v3.14/module/FetchContent.html?highlight=fetchcontent#command:fetchcontent_declare
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs=1,
      kwargs={
          # Download Step Options
          "DOWNLOAD_COMMAND": ShellCommandNode.parse,
          "URL": PositionalParser('+'),
          "URL_HASH": PositionalParser(1),
          "URL_MD5": PositionalParser(1),
          "DOWNLOAD_NAME": PositionalParser(1),
          "DOWNLOAD_NO_EXTRACT": PositionalParser(1),
          "DOWNLOAD_NO_PROGRESS": PositionalParser(1),
          "TIMEOUT": PositionalParser(1),
          "HTTP_USERNAME": PositionalParser(1),
          "HTTP_PASSWORD": PositionalParser(1),
          "HTTP_HEADER": PositionalParser('+'),
          "TLS_VERIFY": PositionalParser(1),
          "TLS_CAINFO": PositionalParser(1),
          "NETRC": PositionalParser(
              1, flags=["CMAKE_NETRC", "IGNORED", "OPTIONAL", "REQUIRED"]),
          "NETRC_FILE": PositionalParser(1),
          # Git
          "GIT_REPOSITORY": PositionalParser(1),
          "GIT_TAG": PositionalParser(1),
          "GIT_REMOTE_NAME": PositionalParser(1),
          "GIT_SUBMODULES": PositionalParser('+'),
          "GIT_SHALLOW": PositionalParser(1),
          "GIT_PROGRESS": PositionalParser(1),
          "GIT_CONFIG": PositionalParser('+'),
          # Subversion
          "SVN_REPOSITORY": PositionalParser(1),
          "SVN_REVISION": PositionalParser(1),
          "SVN_USERNAME": PositionalParser(1),
          "SVN_PASSWORD": PositionalParser(1),
          "SVN_TRUST_CERT": PositionalParser(1),
          # Mercurial
          "HG_REPOSITORY": PositionalParser(1),
          "HG_TAG": PositionalParser(1),
          # CVS
          "CVS_REPOSITORY": PositionalParser(1),
          "CVS_MODULE": PositionalParser(1),
          "CVS_TAG": PositionalParser(1),
          # Update/Patch Step Options
          "UPDATE_COMMAND": ShellCommandNode.parse,
          "UPDATE_DISCONNECTED": PositionalParser(1),
          "PATCH_COMMAND": ShellCommandNode.parse,
      },
      flags=[],
      breakstack=breakstack)


def parse_fetchcontent_populate(ctx, tokens, breakstack):
  """
  ::

    FetchContent_Populate( <name>
      [QUIET]
      [SUBBUILD_DIR <subBuildDir>]
      [SOURCE_DIR <srcDir>]
      [BINARY_DIR <binDir>]
      ...
    )

  :see: https://cmake.org/cmake/help/v3.14/module/FetchContent.html?highlight=fetchcontent#command:fetchcontent_populate
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs=1,
      kwargs={
          "SUBBUILD_DIR": PositionalParser(1),
          "SOURCE_DIR": PositionalParser(1),
          "BINARY_DIR": PositionalParser(1),
          # Download Step Options
          "DOWNLOAD_COMMAND": ShellCommandNode.parse,
          "URL": PositionalParser('+'),
          "URL_HASH": PositionalParser(1),
          "URL_MD5": PositionalParser(1),
          "DOWNLOAD_NAME": PositionalParser(1),
          "DOWNLOAD_NO_EXTRACT": PositionalParser(1),
          "DOWNLOAD_NO_PROGRESS": PositionalParser(1),
          "TIMEOUT": PositionalParser(1),
          "HTTP_USERNAME": PositionalParser(1),
          "HTTP_PASSWORD": PositionalParser(1),
          "HTTP_HEADER": PositionalParser('+'),
          "TLS_VERIFY": PositionalParser(1),
          "TLS_CAINFO": PositionalParser(1),
          "NETRC": PositionalParser(
              1, flags=["CMAKE_NETRC", "IGNORED", "OPTIONAL", "REQUIRED"]),
          "NETRC_FILE": PositionalParser(1),
          # Git
          "GIT_REPOSITORY": PositionalParser(1),
          "GIT_TAG": PositionalParser(1),
          "GIT_REMOTE_NAME": PositionalParser(1),
          "GIT_SUBMODULES": PositionalParser('+'),
          "GIT_SHALLOW": PositionalParser(1),
          "GIT_PROGRESS": PositionalParser(1),
          "GIT_CONFIG": PositionalParser('+'),
          # Subversion
          "SVN_REPOSITORY": PositionalParser(1),
          "SVN_REVISION": PositionalParser(1),
          "SVN_USERNAME": PositionalParser(1),
          "SVN_PASSWORD": PositionalParser(1),
          "SVN_TRUST_CERT": PositionalParser(1),
          # Mercurial
          "HG_REPOSITORY": PositionalParser(1),
          "HG_TAG": PositionalParser(1),
          # CVS
          "CVS_REPOSITORY": PositionalParser(1),
          "CVS_MODULE": PositionalParser(1),
          "CVS_TAG": PositionalParser(1),
          # Update/Patch Step Options
          "UPDATE_COMMAND": ShellCommandNode.parse,
          "UPDATE_DISCONNECTED": PositionalParser(1),
          "PATCH_COMMAND": ShellCommandNode.parse,
      },
      flags=[
          "QUIET"
      ],
      breakstack=breakstack)


def parse_fetchcontent_getproperties(ctx, tokens, breakstack):
  """
  ::

    FetchContent_GetProperties( <name>
      [SOURCE_DIR <srcDirVar>]
      [BINARY_DIR <binDirVar>]
      [POPULATED <doneVar>]
    )
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs=1,
      kwargs={
          "SOURCE_DIR": PositionalParser(1),
          "BINARY_DIR": PositionalParser(1),
          "POPULATED": PositionalParser(1),
      },
      flags=[],
      breakstack=breakstack)


def populate_db(parse_db):
    # Standard, non-builtin commands
  parse_db["fetchcontent_declare"] = parse_fetchcontent_declare
  parse_db["fetchcontent_populate"] = parse_fetchcontent_populate
  parse_db["fetchcontent_getproperties"] = parse_fetchcontent_getproperties
  parse_db["fetchcontent_makeavailable"] = StandardParser('+')



from __future__ import unicode_literals

import logging
import re

from cmakelang.parse import util as parse_util
from cmakelang.parse.funs import standard_funs
from cmakelang import markup
from cmakelang.config_util import (
    FieldDescriptor, ConfigObject, SubtreeDescriptor)


class MarkupConfig(ConfigObject):
  """Options affecting comment reflow and formatting."""

  _field_registry = []

  bullet_char = FieldDescriptor(
      '*',
      "What character to use for bulleted lists"
  )

  enum_char = FieldDescriptor(
      '.',
      "What character to use as punctuation after numerals in an"
      " enumerated list"
  )
  first_comment_is_literal = FieldDescriptor(
      False,
      "If comment markup is enabled, don't reflow the first comment block"
      " in each listfile. Use this to preserve formatting of your"
      " copyright/license statements. "
  )
  literal_comment_pattern = FieldDescriptor(
      None,
      "If comment markup is enabled, don't reflow any comment block which"
      " matches this (regex) pattern. Default is `None` (disabled).",
  )
  fence_pattern = FieldDescriptor(
      markup.FENCE_PATTERN,
      "Regular expression to match preformat fences in comments"
      " default= ``r'{}'``".format(markup.FENCE_PATTERN)
  )
  ruler_pattern = FieldDescriptor(
      markup.RULER_PATTERN,
      "Regular expression to match rulers in comments default= ``r'{}'``"
      .format(markup.RULER_PATTERN)
  )
  explicit_trailing_pattern = FieldDescriptor(
      "#<",
      "If a comment line matches starts with this pattern then it is "
      "explicitly a trailing comment for the preceeding argument. Default "
      "is '#<'"
  )
  hashruler_min_length = FieldDescriptor(
      10,
      "If a comment line starts with at least this many consecutive hash "
      "characters, then don't lstrip() them off. This allows for lazy hash "
      "rulers where the first hash char is not separated by space"
  )
  canonicalize_hashrulers = FieldDescriptor(
      True,
      "If true, then insert a space between the first hash char and"
      " remaining hash chars in a hash ruler, and normalize its length to"
      " fill the column"
  )
  enable_markup = FieldDescriptor(
      True,
      "enable comment markup parsing and reflow"
  )


class EncodingConfig(ConfigObject):
  """Options affecting file encoding"""

  _field_registry = []

  emit_byteorder_mark = FieldDescriptor(
      False,
      "If true, emit the unicode byte-order mark (BOM) at the start of"
      " the file"
  )
  input_encoding = FieldDescriptor(
      "utf-8",
      "Specify the encoding of the input file. Defaults to utf-8"
  )
  output_encoding = FieldDescriptor(
      "utf-8",
      "Specify the encoding of the output file. Defaults to utf-8. Note"
      " that cmake only claims to support utf-8 so be careful when using"
      " anything else"
  )


class LinterConfig(ConfigObject):
  """Options affecting the linter"""

  _field_registry = []

  disabled_codes = FieldDescriptor(
      [],
      "a list of lint codes to disable",
  )

  function_pattern = FieldDescriptor(
      "[0-9a-z_]+",
      "regular expression pattern describing valid function names"
  )

  macro_pattern = FieldDescriptor(
      "[0-9A-Z_]+",
      "regular expression pattern describing valid macro names"
  )

  global_var_pattern = FieldDescriptor(
      "[A-Z][0-9A-Z_]+",
      "regular expression pattern describing valid names for variables"
      " with global (cache) scope"
  )

  internal_var_pattern = FieldDescriptor(
      "_[A-Z][0-9A-Z_]+",
      "regular expression pattern describing valid names for variables"
      " with global scope (but internal semantic)"
  )

  local_var_pattern = FieldDescriptor(
      "[a-z][a-z0-9_]+",
      "regular expression pattern describing valid names for variables"
      " with local scope"
  )

  private_var_pattern = FieldDescriptor(
      "_[0-9a-z_]+",
      "regular expression pattern describing valid names for private"
      "directory variables"
  )

  public_var_pattern = FieldDescriptor(
      "[A-Z][0-9A-Z_]+",
      "regular expression pattern describing valid names for public"
      " directory variables"
  )

  argument_var_pattern = FieldDescriptor(
      "[a-z][a-z0-9_]+",
      "regular expression pattern describing valid names for function/macro"
      " arguments and loop variables."
  )

  keyword_pattern = FieldDescriptor(
      "[A-Z][0-9A-Z_]+",
      "regular expression pattern describing valid names for keywords"
      " used in functions or macros"
  )

  max_conditionals_custom_parser = FieldDescriptor(
      2,
      "In the heuristic for C0201, how many conditionals to match within"
      " a loop in before considering the loop a parser."
  )

  min_statement_spacing = FieldDescriptor(
      1,
      "Require at least this many newlines between statements"
  )

  max_statement_spacing = FieldDescriptor(
      2,
      "Require no more than this many newlines between statements"
  )

  max_returns = FieldDescriptor(6,)
  max_branches = FieldDescriptor(12,)
  max_arguments = FieldDescriptor(5,)
  max_localvars = FieldDescriptor(15,)
  max_statements = FieldDescriptor(50,)


ADDITIONAL_COMMANDS_DEMO = {
    'foo': {
        'flags': ['BAR', 'BAZ'],
        'kwargs': {
            'HEADERS': '*',
            'SOURCES': '*',
            'DEPENDS': '*'
        }
    }
}


class FormattingConfig(ConfigObject):
  """Options affecting formatting."""

  _field_registry = []

  disable = FieldDescriptor(
      False,
      "Disable formatting entirely, making cmake-format a no-op"
  )

  line_width = FieldDescriptor(
      80,
      "How wide to allow formatted cmake files"
  )

  tab_size = FieldDescriptor(
      2,
      "How many spaces to tab for indent"
  )

  use_tabchars = FieldDescriptor(
      False,
      "If true, lines are indented using tab characters (utf-8 0x09) instead"
      " of <tab_size> space characters (utf-8 0x20). In cases where the layout"
      " would require a fractional tab character, the behavior of the "
      " fractional indentation is governed by <fractional_tab_policy>"
  )

  fractional_tab_policy = FieldDescriptor(
      "use-space",
      "If <use_tabchars> is True, then the value of this variable indicates"
      " how fractional indentions are handled during whitespace replacement."
      " If set to 'use-space', fractional indentation is left as spaces"
      " (utf-8 0x20). If set to `round-up` fractional indentation is replaced"
      " with a single tab character (utf-8 0x09) effectively shifting the"
      " column to the next tabstop",
      ["use-space", "round-up"]
  )

  max_subgroups_hwrap = FieldDescriptor(
      2,
      "If an argument group contains more than this many sub-groups "
      "(parg or kwarg groups) then force it to a vertical layout. "
  )

  max_pargs_hwrap = FieldDescriptor(
      6,
      "If a positional argument group contains more than this many"
      " arguments, then force it to a vertical layout. "
  )

  max_rows_cmdline = FieldDescriptor(
      2,
      "If a cmdline positional group consumes more than this many lines"
      " without nesting, then invalidate the layout (and nest)"
  )

  separate_ctrl_name_with_space = FieldDescriptor(
      False,
      "If true, separate flow control names from their parentheses with a"
      " space"
  )

  separate_fn_name_with_space = FieldDescriptor(
      False,
      "If true, separate function names from parentheses with a space"
  )

  dangle_parens = FieldDescriptor(
      False,
      "If a statement is wrapped to more than one line, than dangle the"
      " closing parenthesis on its own line."
  )

  dangle_align = FieldDescriptor(
      "prefix",
      "If the trailing parenthesis must be 'dangled' on its on line, then"
      " align it to this reference: `prefix`: the start of the statement, "
      " `prefix-indent`: the start of the statement, plus one indentation "
      " level, `child`: align to the column of the arguments",
      ["prefix", "prefix-indent", "child", "off"],
  )

  min_prefix_chars = FieldDescriptor(
      4,
      "If the statement spelling length (including space and parenthesis)"
      " is smaller than this amount, then force reject nested layouts."
  )

  max_prefix_chars = FieldDescriptor(
      10,
      "If the statement spelling length (including space and parenthesis)"
      " is larger than the tab width by more than this amount, then"
      " force reject un-nested layouts."
  )

  max_lines_hwrap = FieldDescriptor(
      2,
      "If a candidate layout is wrapped horizontally but it exceeds this"
      " many lines, then reject the layout."
  )

  line_ending = FieldDescriptor(
      "unix",
      "What style line endings to use in the output.",
      ['windows', 'unix', 'auto']
  )

  command_case = FieldDescriptor(
      "canonical",
      "Format command names consistently as 'lower' or 'upper' case",
      ['lower', 'upper', 'canonical', 'unchanged']
  )

  keyword_case = FieldDescriptor(
      "unchanged",
      "Format keywords consistently as 'lower' or 'upper' case",
      ['lower', 'upper', 'unchanged']
  )

  always_wrap = FieldDescriptor(
      [],
      "A list of command names which should always be wrapped"
  )

  enable_sort = FieldDescriptor(
      True,
      "If true, the argument lists which are known to be sortable will be "
      "sorted lexicographicall"
  )

  autosort = FieldDescriptor(
      False,
      "If true, the parsers may infer whether or not an argument list is"
      " sortable (without annotation)."
  )

  require_valid_layout = FieldDescriptor(
      False,
      "By default, if cmake-format cannot successfully fit everything into"
      " the desired linewidth it will apply the last, most agressive"
      " attempt that it made. If this flag is True, however, cmake-format"
      " will print error, exit with non-zero status code, and write-out"
      " nothing"
  )

  layout_passes = FieldDescriptor(
      {},
      "A dictionary mapping layout nodes to a list of wrap decisions. See"
      " the documentation for more information."
  )

  def __init__(self, **kwargs):
    super(FormattingConfig, self).__init__(**kwargs)
    self.endl = None

  @property
  def linewidth(self):
    return self.line_width

  def set_line_ending(self, detected):
    self.endl = {
        "windows": "\r\n",
        "unix": "\n"
    }[detected]

  def _update_derived(self):
    """Update derived values after a potential config change
    """
    self.endl = {
        "windows": "\r\n",
        "unix": "\n",
        "auto": "\n"
    }[self.line_ending]


BUILTIN_VARTAGS = [
    (".*_COMMAND", ["cmdline"])
]

BUILTIN_PROPTAGS = [
    (".*_DIRECTORIES", ["file-list"])
]


class ParseConfig(ConfigObject):
  """Options affecting listfile parsing"""

  _field_registry = []

  additional_commands = FieldDescriptor(
      ADDITIONAL_COMMANDS_DEMO,
      "Specify structure for custom cmake functions"
  )

  override_spec = FieldDescriptor(
      {},
      "Override configurations per-command where available")

  vartags = FieldDescriptor([], "Specify variable tags.")
  proptags = FieldDescriptor([], "Specify property tags.")

  def __init__(self, **kwargs):  # pylint: disable=W0613
    self.fn_spec = parse_util.CommandSpec("<root>")
    self.vartags_ = []
    self.proptags_ = []
    super(ParseConfig, self).__init__(**kwargs)

  def _update_derived(self):
    if self.additional_commands is not None:
      for command_name, spec in self.additional_commands.items():
        self.fn_spec.add(command_name, **spec)

    for pathkeystr, value in self.override_spec.items():
      parse_util.apply_overrides(self.fn_spec, pathkeystr, value)

    self.vartags_ = [
        (re.compile(pattern, re.IGNORECASE), tags) for pattern, tags in
        BUILTIN_VARTAGS + self.vartags]
    self.proptags_ = [
        (re.compile(pattern, re.IGNORECASE), tags) for pattern, tags in
        BUILTIN_PROPTAGS + self.proptags]


class MiscConfig(ConfigObject):
  """Miscellaneous configurations options."""

  _field_registry = []

  per_command = FieldDescriptor(
      {},
      "A dictionary containing any per-command configuration overrides."
      " Currently only `command_case` is supported."
  )

  def _update_derived(self):
    self.per_command_ = standard_funs.get_default_config()
    for command, cdict in self.per_command.items():
      if not isinstance(cdict, dict):
        logging.warning("Invalid override of type %s for %s",
                        type(cdict), command)
        continue

      command = command.lower()
      if command not in self.per_command_:
        self.per_command_[command] = {}
      self.per_command_[command].update(cdict)

  def __init__(self, **kwargs):  # pylint: disable=W0613
    self.per_command_ = {}
    super(MiscConfig, self).__init__(**kwargs)


class Configuration(ConfigObject):
  """Various configuration options and parameters"""
  _field_registry = []

  parse = SubtreeDescriptor(ParseConfig)
  format = SubtreeDescriptor(FormattingConfig)
  markup = SubtreeDescriptor(MarkupConfig)
  lint = SubtreeDescriptor(LinterConfig)
  encode = SubtreeDescriptor(EncodingConfig)
  misc = SubtreeDescriptor(MiscConfig)

  def resolve_for_command(self, command_name, config_key, default_value=None):
    """
    Check for a per-command value or override of the given configuration key
    and return it if it exists. Otherwise return the global configuration value
    for that key.
    """

    configpath = config_key.split(".")
    fieldname = configpath.pop(-1)
    configobj = self
    for subname in configpath:
      nextobj = getattr(configobj, subname, None)
      if nextobj is None:
        raise ValueError(
            "Config object {} does not have a subobject named {}"
            .format(type(configobj).__name__, subname))
      configobj = nextobj

    if hasattr(configobj, fieldname):
      assert default_value is None, (
          "Specifying a default value is not allowed if the config key exists "
          "in the global configuration ({})".format(config_key))
      default_value = getattr(configobj, fieldname)

    command_dict = self.misc.per_command_.get(command_name.lower(), {})
    if config_key in command_dict:
      return command_dict[config_key]

    # legacy unqualified fieldname
    if fieldname in command_dict:
      return command_dict[fieldname]

    return default_value
